import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Description, Name, NamedResource

from .utils import PokemonEntry

__all__: t.Tuple[str, ...] = (
    "Generation",
    "Pokedex",
    "Version",
    "VersionGroup",
)


@attrs.define(slots=True, kw_only=True)
class Generation(BaseModel):
    """Generation model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    abilities: t.List[NamedResource]
        A list of abilities that were introduced in this generation.
    names: t.List[Name]
        The name of this resource listed in different languages.
    main_region: NamedResource
        The main region travelled in this generation.
    moves: t.List[NamedResource]
        A list of moves that were introduced in this generation.
    pokemon_species: t.List[NamedResource]
        A list of Pokémon species that were introduced in this generation.
    types: t.List[NamedResource]
        A list of types that were introduced in this generation.
    version_groups: t.List[NamedResource]
        A list of version groups that were introduced in this generation.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    abilities: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    main_region: NamedResource = attrs.field(factory=NamedResource)
    moves: t.List[NamedResource] = attrs.field(factory=list)
    pokemon_species: t.List[NamedResource] = attrs.field(factory=list)
    types: t.List[NamedResource] = attrs.field(factory=list)
    version_groups: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Generation":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            abilities=[NamedResource.from_payload(i) for i in payload.get("abilities", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            main_region=NamedResource.from_payload(payload.get("main_region", {}) or {}),
            moves=[NamedResource.from_payload(i) for i in payload.get("moves", [])],
            pokemon_species=[NamedResource.from_payload(i) for i in payload.get("pokemon_species", [])],
            types=[NamedResource.from_payload(i) for i in payload.get("types", [])],
            version_groups=[NamedResource.from_payload(i) for i in payload.get("version_groups", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Pokedex(BaseModel):
    """Pokedex model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    is_main_series: bool
        Whether or not this Pokédex originated in the main series of the video games.
    descriptions: t.List[Description]
        The description of this resource listed in different languages.
    names: t.List[Name]
        The name of this resource listed in different languages.
    pokemon_entries: t.List[PokemonEntry]
        A list of Pokémon catalogued in this Pokédex and their indexes.
    region: NamedResource
        The region this Pokédex catalogues Pokémon for.
    version_groups: t.List[NamedResource]
        A list of version groups this Pokédex is relevant to.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    is_main_series: bool = attrs.field(factory=bool)
    descriptions: t.List[Description] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)
    pokemon_entries: t.List[PokemonEntry] = attrs.field(factory=list)
    region: NamedResource = attrs.field(factory=NamedResource)
    version_groups: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Pokedex":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            is_main_series=payload.get("is_main_series", False),
            descriptions=[Description.from_payload(i) for i in payload.get("descriptions", [])],
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            pokemon_entries=[PokemonEntry.from_payload(i) for i in payload.get("pokemon_entries", [])],
            region=NamedResource.from_payload(payload.get("region", {}) or {}),
            version_groups=[NamedResource.from_payload(i) for i in payload.get("version_groups", [])],
        )


@attrs.define(slots=True, kw_only=True)
class Version(BaseModel):
    """Version model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    names: t.List[Name]
        The name of this resource listed in different languages.
    version_group: NamedResource
        The version group this version belongs to.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Version":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class VersionGroup(BaseModel):
    """VersionGroup model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    order: int
        Order for sorting. Almost by date of release, except similar versions are grouped together.
    generation: NamedResource
        The generation this version was introduced in.
    move_learn_methods: t.List[NamedResource]
        A list of methods in which Pokémon can learn moves in this version group.
    pokedexes: t.List[NamedResource]
        A list of Pokédexes introduces in this version group.
    regions: t.List[NamedResource]
        A list of regions that can be visited in this version group.
    versions: t.List[NamedResource]
        A list of versions this version group owns.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    order: int = attrs.field(factory=int)
    generation: NamedResource = attrs.field(factory=NamedResource)
    move_learn_methods: t.List[NamedResource] = attrs.field(factory=list)
    pokedexes: t.List[NamedResource] = attrs.field(factory=list)
    regions: t.List[NamedResource] = attrs.field(factory=list)
    versions: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "VersionGroup":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            order=payload.get("order", 0),
            generation=NamedResource.from_payload(payload.get("generation", {}) or {}),
            move_learn_methods=[NamedResource.from_payload(i) for i in payload.get("move_learn_methods", [])],
            pokedexes=[NamedResource.from_payload(i) for i in payload.get("pokedexes", [])],
            regions=[NamedResource.from_payload(i) for i in payload.get("regions", [])],
            versions=[NamedResource.from_payload(i) for i in payload.get("versions", [])],
        )



import datetime
import enum
import logging
import pathlib
import typing as t

__all__: t.Tuple[str, ...] = (
    "Logger",
    "FileHandler",
    "Formatter",
)
FLAIR: int = 95


class LogLevelColors(enum.Enum):
    """Colors for the log levels."""

    DEBUG = "\033[96m"
    INFO = "\033[92m"
    WARNING = "\033[93m"
    ERROR = "\033[33m"
    CRITICAL = "\033[91m"
    ENDC = "\033[0m"
    FLAIR = "\033[95m"


class Formatter(logging.Formatter):
    """Format the log record."""

    def __init__(self) -> None:
        super().__init__(
            "[%(asctime)s] | %(pathname)s:%(lineno)d | %(levelname)s | %(message)s",
            style="%",
        )

    def format(self, record: logging.LogRecord) -> str:
        """Format the log record."""
        return f"{LogLevelColors[record.levelname].value}{super().format(record)}{LogLevelColors.ENDC.value}"


class FileHandler(logging.FileHandler):
    """Emit a log record.

    Parameters
    ----------
    ext : str
        The file extension.
    folder : pathlib.Path | str
        The folder to save the logs in. Defaults to "logs".
    """

    _last_entry: datetime.datetime = datetime.datetime.today()

    def __init__(self, *, ext: str, folder: t.Union[pathlib.Path, str] = "logs") -> None:
        """Create a new file handler."""
        self.folder = pathlib.Path(folder)
        self.ext = ext
        self.folder.mkdir(exist_ok=True)
        super().__init__(
            self.folder / f"{datetime.datetime.today().strftime('%Y-%m-%d')}-{ext}.log",
            encoding="utf-8",
        )
        self.setFormatter(Formatter())

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a log record."""
        if self._last_entry.date() != datetime.datetime.today().date():
            self._last_entry = datetime.datetime.today()
            self.close()
            self.baseFilename = (self.folder / f"{self._last_entry.strftime('%Y-%m-%d')}-{self.ext}.log").as_posix()
            self.stream = self._open()
        super().emit(record)


class Logger(logging.Logger):
    """
    The logger used to log information about the client.

    Parameters
    ----------
    name : str
        The name of the logger.
    level : int
        The level of the logger.
    file_logging : bool
        Whether or not to log to a file.

    Attributes
    ----------
    _handler : logging.StreamHandler
        The stream handler used to log to the console.
    _file_handler : t.Optional[logging.FileHandler]
        The file handler used to log to a file.

    Examples
    --------

    >>> logs = Logger(name="pokelance")
    >>> logs.info("Hello, world!")
    [2021-08-29 17:05:32,000] | pokelance/logger.py:95 | INFO | Hello, world!
    """

    file_handler: t.Optional[FileHandler] = None

    def __init__(self, *, name: str, level: int = logging.DEBUG, file_logging: bool = False) -> None:
        super().__init__(name, level)
        self._handler = logging.StreamHandler()
        self._handler.setFormatter(Formatter())
        self.addHandler(self._handler)
        if file_logging:
            self._file_handler = FileHandler(ext=name)
            self.addHandler(self._file_handler)
        logging.addLevelName(FLAIR, "FLAIR")

    def set_formatter(self, formatter: logging.Formatter) -> None:
        """Set the formatter."""
        self._handler.setFormatter(formatter)
        self._file_handler.setFormatter(formatter)

    def flair(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """Record a flair log."""
        self.log(FLAIR, message, *args, **kwargs)



import typing as t

from pokelance.http import Endpoint
from pokelance.models import EvolutionChain, EvolutionTrigger

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.cache import Evolution as EvolutionCache


__all__: t.Tuple[str, ...] = (
    "setup",
    "Evolution",
)


class Evolution(BaseExtension):
    """
    Extension for evolution related endpoints.

    Attributes
    ----------
    cache: pokelance.cache.Evolution
        The cache for this extension.
    """

    cache: "EvolutionCache"

    def get_evolution_chain(self, id_: int) -> t.Optional[EvolutionChain]:
        """Gets an evolution chain from the cache.

        Parameters
        ----------
        id_: int
            The name or id of the encounter method.

        Returns
        -------
        typing.Optional[pokelance.models.EvolutionChain]
            The evolution chain if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the evolution chain is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> chain = client.evolution.get_evolution_chain(1)
        >>> chain.id
        1
        """
        route = Endpoint.get_evolution_chain(id_)
        self._validate_resource(self.cache.evolution_chain, id_, route)
        return self.cache.evolution_chain.get(route, None)

    async def fetch_evolution_chain(self, id_: int) -> EvolutionChain:
        """Fetches an evolution chain from the API.

        Parameters
        ----------
        id_: int
            The name or id of the encounter method.

        Returns
        -------
        EvolutionChain
            The evolution chain if it exists in the cache, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the evolution chain is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     chain = await client.evolution.fetch_evolution_chain(1)
        ...     print(chain.id)
        ...     await client.close()
        >>> asyncio.run(main())
        1
        """
        route = Endpoint.get_evolution_chain(id_)
        self._validate_resource(self.cache.evolution_chain, id_, route)
        data = await self._client.request(route)
        return self.cache.evolution_chain.setdefault(route, EvolutionChain.from_payload(data))

    def get_evolution_trigger(self, name: t.Union[str, int]) -> t.Optional[EvolutionTrigger]:
        """Gets an evolution trigger from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter method.

        Returns
        -------
        typing.Optional[pokelance.models.EvolutionTrigger]
            The evolution trigger if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the evolution trigger is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> trigger = client.evolution.get_evolution_trigger(1)
        >>> trigger.name
        'level-up'
        """
        route = Endpoint.get_evolution_trigger(name)
        self._validate_resource(self.cache.evolution_trigger, name, route)
        return self.cache.evolution_trigger.get(route, None)

    async def fetch_evolution_trigger(self, name: t.Union[str, int]) -> EvolutionTrigger:
        """Fetches an evolution trigger from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the encounter method.

        Returns
        -------
        EvolutionTrigger
            The evolution trigger if it exists in the API, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the evolution trigger is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     trigger = await client.evolution.fetch_evolution_trigger(1)
        ...     print(trigger.name)
        ...     await client.close()
        >>> asyncio.run(main())
        level-up
        """
        route = Endpoint.get_evolution_trigger(name)
        self._validate_resource(self.cache.evolution_trigger, name, route)
        data = await self._client.request(route)
        return self.cache.evolution_trigger.setdefault(route, EvolutionTrigger.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the evolution cog."""
    lance.add_extension("evolution", Evolution(lance.http))



import typing as t

import attrs

__all__: t.Tuple[str, ...] = ("Endpoint", "Route")


@attrs.define(repr=True, slots=True, kw_only=True, hash=True)
class Route:
    """Represents a route for an endpoint.

    Attributes
    ----------
    endpoint: str
        The endpoint name.
    _url: str
        The URL for the route.
    _api_version: int
        The API version for the route.
    method: str
        The HTTP method for the route.
    payload: t.Optional[t.Dict[str, t.Any]]
        The payload for the route.

    Examples
    --------
    >>> pokemon = "pikachu"
    >>> route = Route(endpoint=f"/pokemon/{pokemon}", method="GET")
    """

    endpoint: str = attrs.field(factory=str)
    _url: str = "https://pokeapi.co/api/v2{endpoint}"
    _api_version: int = 2
    method: str = "GET"
    payload: t.Optional[t.Dict[str, t.Any]] = None

    def __str__(self) -> str:
        return f"<Route endpoint={self.endpoint} method={self.method}>"

    @property
    def url(self) -> str:
        return self._url.format(endpoint=self.endpoint)


class Endpoint:
    """Represents an endpoint for the API."""

    @classmethod
    def get_language_endpoints(cls) -> Route:
        """Gets the language endpoints."""
        return Route(endpoint="/language", payload={"limit": 10000})

    @classmethod
    def get_language(cls, language: t.Union[str, int]) -> Route:
        """Gets a language."""
        return Route(endpoint=f"/language/{language}")

    @classmethod
    def get_berry_endpoints(cls) -> Route:
        """Get a list of berry endpoints."""
        return Route(endpoint="/berry", payload={"limit": 10000})

    @classmethod
    def get_berry(cls, berry: t.Union[int, str]) -> Route:
        """Get a berry by its ID or name."""
        return Route(endpoint=f"/berry/{berry}")

    @classmethod
    def get_berry_firmness_endpoints(cls) -> Route:
        """Get a list of berry firmness endpoints."""
        return Route(endpoint="/berry-firmness", payload={"limit": 10000})

    @classmethod
    def get_berry_firmness(cls, berry_firmness: t.Union[int, str]) -> Route:
        """Get a berry firmness by its ID or name."""
        return Route(endpoint=f"/berry-firmness/{berry_firmness}")

    @classmethod
    def get_berry_flavor_endpoints(cls) -> Route:
        """Get a list of berry flavor endpoints."""
        return Route(endpoint="/berry-flavor", payload={"limit": 10000})

    @classmethod
    def get_berry_flavor(cls, berry_flavor: t.Union[int, str]) -> Route:
        """Get a berry flavor by its ID or name."""
        return Route(endpoint=f"/berry-flavor/{berry_flavor}")

    @classmethod
    def get_contest_type_endpoints(cls) -> Route:
        """Get a list of contest type endpoints."""
        return Route(endpoint="/contest-type", payload={"limit": 10000})

    @classmethod
    def get_contest_type(cls, contest_type: t.Union[int, str]) -> Route:
        """Get a contest type by its ID or name."""
        return Route(endpoint=f"/contest-type/{contest_type}")

    @classmethod
    def get_contest_effect_endpoints(cls) -> Route:
        """Get a list of contest effect endpoints."""
        return Route(endpoint="/contest-effect", payload={"limit": 10000})

    @classmethod
    def get_contest_effect(cls, contest_effect: int) -> Route:
        """Get a contest effect by its ID."""
        return Route(endpoint=f"/contest-effect/{contest_effect}")

    @classmethod
    def get_super_contest_effect_endpoints(cls) -> Route:
        """Get a list of super contest effect endpoints."""
        return Route(endpoint="/super-contest-effect", payload={"limit": 10000})

    @classmethod
    def get_super_contest_effect(cls, super_contest_effect: int) -> Route:
        """Get a super contest effect by its ID."""
        return Route(endpoint=f"/super-contest-effect/{super_contest_effect}")

    @classmethod
    def get_encounter_method_endpoints(cls) -> Route:
        """Get a list of encounter method endpoints."""
        return Route(endpoint="/encounter-method", payload={"limit": 10000})

    @classmethod
    def get_encounter_method(cls, encounter_method: t.Union[int, str]) -> Route:
        """Get an encounter method by its ID or name."""
        return Route(endpoint=f"/encounter-method/{encounter_method}")

    @classmethod
    def get_encounter_condition_endpoints(cls) -> Route:
        """Get a list of encounter condition endpoints."""
        return Route(endpoint="/encounter-condition", payload={"limit": 10000})

    @classmethod
    def get_encounter_condition(cls, encounter_condition: t.Union[int, str]) -> Route:
        """Get an encounter condition by its ID or name."""
        return Route(endpoint=f"/encounter-condition/{encounter_condition}")

    @classmethod
    def get_encounter_condition_value_endpoints(cls) -> Route:
        """Get a list of encounter condition value endpoints."""
        return Route(endpoint="/encounter-condition-value", payload={"limit": 10000})

    @classmethod
    def get_encounter_condition_value(cls, encounter_condition_value: t.Union[int, str]) -> Route:
        """Get an encounter condition value by its ID or name."""
        return Route(endpoint=f"/encounter-condition-value/{encounter_condition_value}")

    @classmethod
    def get_evolution_chain_endpoints(cls) -> Route:
        """Get a list of evolution chain endpoints."""
        return Route(endpoint="/evolution-chain", payload={"limit": 10000})

    @classmethod
    def get_evolution_chain(cls, evolution_chain: int) -> Route:
        """Get an evolution chain by its ID."""
        return Route(endpoint=f"/evolution-chain/{evolution_chain}")

    @classmethod
    def get_evolution_trigger_endpoints(cls) -> Route:
        """Get a list of evolution trigger endpoints."""
        return Route(endpoint="/evolution-trigger", payload={"limit": 10000})

    @classmethod
    def get_evolution_trigger(cls, evolution_trigger: t.Union[int, str]) -> Route:
        """Get an evolution trigger by its ID or name."""
        return Route(endpoint=f"/evolution-trigger/{evolution_trigger}")

    @classmethod
    def get_generation_endpoints(cls) -> Route:
        """Get a list of generation endpoints."""
        return Route(endpoint="/generation", payload={"limit": 10000})

    @classmethod
    def get_generation(cls, generation: t.Union[int, str]) -> Route:
        """Get a generation by its ID or name."""
        return Route(endpoint=f"/generation/{generation}")

    @classmethod
    def get_pokedex_endpoints(cls) -> Route:
        """Get a list of pokedex endpoints."""
        return Route(endpoint="/pokedex", payload={"limit": 10000})

    @classmethod
    def get_pokedex(cls, pokedex: t.Union[int, str]) -> Route:
        """Get a pokedex by its ID or name."""
        return Route(endpoint=f"/pokedex/{pokedex}")

    @classmethod
    def get_version_endpoints(cls) -> Route:
        """Get a list of version endpoints."""
        return Route(endpoint="/version", payload={"limit": 10000})

    @classmethod
    def get_version(cls, version: t.Union[int, str]) -> Route:
        """Get a version by its ID or name."""
        return Route(endpoint=f"/version/{version}")

    @classmethod
    def get_version_group_endpoints(cls) -> Route:
        """Get a list of version group endpoints."""
        return Route(endpoint="/version-group", payload={"limit": 10000})

    @classmethod
    def get_version_group(cls, version_group: t.Union[int, str]) -> Route:
        """Get a version group by its ID or name."""
        return Route(endpoint=f"/version-group/{version_group}")

    @classmethod
    def get_item_endpoints(cls) -> Route:
        """Get a list of item endpoints."""
        return Route(endpoint="/item", payload={"limit": 10000})

    @classmethod
    def get_item(cls, item: t.Union[int, str]) -> Route:
        """Get an item by its ID or name."""
        return Route(endpoint=f"/item/{item}")

    @classmethod
    def get_item_attribute_endpoints(cls) -> Route:
        """Get a list of item attribute endpoints."""
        return Route(endpoint="/item-attribute", payload={"limit": 10000})

    @classmethod
    def get_item_attribute(cls, item_attribute: t.Union[int, str]) -> Route:
        """Get an item attribute by its ID or name."""
        return Route(endpoint=f"/item-attribute/{item_attribute}")

    @classmethod
    def get_item_category_endpoints(cls) -> Route:
        """Get a list of item category endpoints."""
        return Route(endpoint="/item-category", payload={"limit": 10000})

    @classmethod
    def get_item_category(cls, item_category: t.Union[int, str]) -> Route:
        """Get an item category by its ID or name."""
        return Route(endpoint=f"/item-category/{item_category}")

    @classmethod
    def get_item_fling_effect_endpoints(cls) -> Route:
        """Get a list of item fling effect endpoints."""
        return Route(endpoint="/item-fling-effect", payload={"limit": 10000})

    @classmethod
    def get_item_fling_effect(cls, item_fling_effect: t.Union[int, str]) -> Route:
        """Get an item fling effect by its ID or name."""
        return Route(endpoint=f"/item-fling-effect/{item_fling_effect}")

    @classmethod
    def get_item_pocket_endpoints(cls) -> Route:
        """Get a list of item pocket endpoints."""
        return Route(endpoint="/item-pocket", payload={"limit": 10000})

    @classmethod
    def get_item_pocket(cls, item_pocket: t.Union[int, str]) -> Route:
        """Get an item pocket by its ID or name."""
        return Route(endpoint=f"/item-pocket/{item_pocket}")

    @classmethod
    def get_location_endpoints(cls) -> Route:
        """Get a list of location endpoints."""
        return Route(endpoint="/location", payload={"limit": 10000})

    @classmethod
    def get_location(cls, location: t.Union[int, str]) -> Route:
        """Get a location by its ID or name."""
        return Route(endpoint=f"/location/{location}")

    @classmethod
    def get_location_area_endpoints(cls) -> Route:
        """Get a list of location area endpoints."""
        return Route(endpoint="/location-area", payload={"limit": 10000})

    @classmethod
    def get_location_area(cls, location_area: t.Union[int, str]) -> Route:
        """Get a location area by its ID or name."""
        return Route(endpoint=f"/location-area/{location_area}")

    @classmethod
    def get_pal_park_area_endpoints(cls) -> Route:
        """Get a list of pal park area endpoints."""
        return Route(endpoint="/pal-park-area", payload={"limit": 10000})

    @classmethod
    def get_pal_park_area(cls, pal_park_area: t.Union[int, str]) -> Route:
        """Get a pal park area by its ID or name."""
        return Route(endpoint=f"/pal-park-area/{pal_park_area}")

    @classmethod
    def get_region_endpoints(cls) -> Route:
        """Get a list of region endpoints."""
        return Route(endpoint="/region", payload={"limit": 10000})

    @classmethod
    def get_region(cls, region: t.Union[int, str]) -> Route:
        """Get a region by its ID or name."""
        return Route(endpoint=f"/region/{region}")

    @classmethod
    def get_machine_endpoints(cls) -> Route:
        """Get a list of machine endpoints."""
        return Route(endpoint="/machine", payload={"limit": 10000})

    @classmethod
    def get_machine(cls, machine: int) -> Route:
        """Get a machine by its ID."""
        return Route(endpoint=f"/machine/{machine}")

    @classmethod
    def get_move_endpoints(cls) -> Route:
        """Get a list of move endpoints."""
        return Route(endpoint="/move", payload={"limit": 10000})

    @classmethod
    def get_move(cls, move: t.Union[int, str]) -> Route:
        """Get a move by its ID or name."""
        return Route(endpoint=f"/move/{move}")

    @classmethod
    def get_move_ailment_endpoints(cls) -> Route:
        """Get a list of move ailment endpoints."""
        return Route(endpoint="/move-ailment", payload={"limit": 10000})

    @classmethod
    def get_move_ailment(cls, move_ailment: t.Union[int, str]) -> Route:
        """Get a move ailment by its ID or name."""
        return Route(endpoint=f"/move-ailment/{move_ailment}")

    @classmethod
    def get_move_battle_style_endpoints(cls) -> Route:
        """Get a list of move battle style endpoints."""
        return Route(endpoint="/move-battle-style", payload={"limit": 10000})

    @classmethod
    def get_move_battle_style(cls, move_battle_style: t.Union[int, str]) -> Route:
        """Get a move battle style by its ID or name."""
        return Route(endpoint=f"/move-battle-style/{move_battle_style}")

    @classmethod
    def get_move_category_endpoints(cls) -> Route:
        """Get a list of move category endpoints."""
        return Route(endpoint="/move-category", payload={"limit": 10000})

    @classmethod
    def get_move_category(cls, move_category: t.Union[int, str]) -> Route:
        """Get a move category by its ID or name."""
        return Route(endpoint=f"/move-category/{move_category}")

    @classmethod
    def get_move_damage_class_endpoints(cls) -> Route:
        """Get a list of move damage class endpoints."""
        return Route(endpoint="/move-damage-class", payload={"limit": 10000})

    @classmethod
    def get_move_damage_class(cls, move_damage_class: t.Union[int, str]) -> Route:
        """Get a move damage class by its ID or name."""
        return Route(endpoint=f"/move-damage-class/{move_damage_class}")

    @classmethod
    def get_move_learn_method_endpoints(cls) -> Route:
        """Get a list of move learn method endpoints."""
        return Route(endpoint="/move-learn-method", payload={"limit": 10000})

    @classmethod
    def get_move_learn_method(cls, move_learn_method: t.Union[int, str]) -> Route:
        """Get a move learn method by its ID or name."""
        return Route(endpoint=f"/move-learn-method/{move_learn_method}")

    @classmethod
    def get_move_target_endpoints(cls) -> Route:
        """Get a list of move target endpoints."""
        return Route(endpoint="/move-target", payload={"limit": 10000})

    @classmethod
    def get_move_target(cls, move_target: t.Union[int, str]) -> Route:
        """Get a move target by its ID or name."""
        return Route(endpoint=f"/move-target/{move_target}")

    @classmethod
    def get_ability_endpoints(cls) -> Route:
        """Get a list of ability endpoints."""
        return Route(endpoint="/ability", payload={"limit": 10000})

    @classmethod
    def get_ability(cls, ability: t.Union[int, str]) -> Route:
        """Get an ability by its ID or name."""
        return Route(endpoint=f"/ability/{ability}")

    @classmethod
    def get_characteristic_endpoints(cls) -> Route:
        """Get a list of characteristic endpoints."""
        return Route(endpoint="/characteristic", payload={"limit": 10000})

    @classmethod
    def get_characteristic(cls, characteristic: int) -> Route:
        """Get a characteristic by its ID."""
        return Route(endpoint=f"/characteristic/{characteristic}")

    @classmethod
    def get_egg_group_endpoints(cls) -> Route:
        """Get a list of egg group endpoints."""
        return Route(endpoint="/egg-group", payload={"limit": 10000})

    @classmethod
    def get_egg_group(cls, egg_group: t.Union[int, str]) -> Route:
        """Get an egg group by its ID or name."""
        return Route(endpoint=f"/egg-group/{egg_group}")

    @classmethod
    def get_gender_endpoints(cls) -> Route:
        """Get a list of gender endpoints."""
        return Route(endpoint="/gender", payload={"limit": 10000})

    @classmethod
    def get_gender(cls, gender: t.Union[int, str]) -> Route:
        """Get a gender by its ID or name."""
        return Route(endpoint=f"/gender/{gender}")

    @classmethod
    def get_growth_rate_endpoints(cls) -> Route:
        """Get a list of growth rate endpoints."""
        return Route(endpoint="/growth-rate", payload={"limit": 10000})

    @classmethod
    def get_growth_rate(cls, growth_rate: t.Union[int, str]) -> Route:
        """Get a growth rate by its ID or name."""
        return Route(endpoint=f"/growth-rate/{growth_rate}")

    @classmethod
    def get_nature_endpoints(cls) -> Route:
        """Get a list of nature endpoints."""
        return Route(endpoint="/nature", payload={"limit": 10000})

    @classmethod
    def get_nature(cls, nature: t.Union[int, str]) -> Route:
        """Get a nature by its ID or name."""
        return Route(endpoint=f"/nature/{nature}")

    @classmethod
    def get_location_area_encounter_endpoints(cls) -> Route:
        """Get a list of location area encounter endpoints."""
        return Route(endpoint="/pokemon", payload={"limit": 10000})

    @classmethod
    def get_location_area_encounter(cls, name: t.Union[int, str]) -> Route:
        """Get a location area encounter by its ID or name."""
        return Route(endpoint=f"/pokemon/{name}/encounters")

    @classmethod
    def get_pokeathlon_stat_endpoints(cls) -> Route:
        """Get a list of pokeathlon stat endpoints."""
        return Route(endpoint="/pokeathlon-stat", payload={"limit": 10000})

    @classmethod
    def get_pokeathlon_stat(cls, pokeathlon_stat: t.Union[int, str]) -> Route:
        """Get a pokeathlon stat by its ID or name."""
        return Route(endpoint=f"/pokeathlon-stat/{pokeathlon_stat}")

    @classmethod
    def get_pokemon_endpoints(cls) -> Route:
        """Get a list of pokemon endpoints."""
        return Route(endpoint="/pokemon", payload={"limit": 10000})

    @classmethod
    def get_pokemon(cls, pokemon: t.Union[int, str]) -> Route:
        """Get a pokemon by its ID or name."""
        return Route(endpoint=f"/pokemon/{pokemon}")

    @classmethod
    def get_pokemon_color_endpoints(cls) -> Route:
        """Get a list of pokemon color endpoints."""
        return Route(endpoint="/pokemon-color", payload={"limit": 10000})

    @classmethod
    def get_pokemon_color(cls, pokemon_color: t.Union[int, str]) -> Route:
        """Get a pokemon color by its ID or name."""
        return Route(endpoint=f"/pokemon-color/{pokemon_color}")

    @classmethod
    def get_pokemon_form_endpoints(cls) -> Route:
        """Get a list of pokemon form endpoints."""
        return Route(endpoint="/pokemon-form", payload={"limit": 10000})

    @classmethod
    def get_pokemon_form(cls, pokemon_form: t.Union[int, str]) -> Route:
        """Get a pokemon form by its ID or name."""
        return Route(endpoint=f"/pokemon-form/{pokemon_form}")

    @classmethod
    def get_pokemon_habitat_endpoints(cls) -> Route:
        """Get a list of pokemon habitat endpoints."""
        return Route(endpoint="/pokemon-habitat", payload={"limit": 10000})

    @classmethod
    def get_pokemon_habitat(cls, pokemon_habitat: t.Union[int, str]) -> Route:
        """Get a pokemon habitat by its ID or name."""
        return Route(endpoint=f"/pokemon-habitat/{pokemon_habitat}")

    @classmethod
    def get_pokemon_shape_endpoints(cls) -> Route:
        """Get a list of pokemon shape endpoints."""
        return Route(endpoint="/pokemon-shape", payload={"limit": 10000})

    @classmethod
    def get_pokemon_shape(cls, pokemon_shape: t.Union[int, str]) -> Route:
        """Get a pokemon shape by its ID or name."""
        return Route(endpoint=f"/pokemon-shape/{pokemon_shape}")

    @classmethod
    def get_pokemon_species_endpoints(cls) -> Route:
        """Get a list of pokemon species endpoints."""
        return Route(endpoint="/pokemon-species", payload={"limit": 10000})

    @classmethod
    def get_pokemon_species(cls, pokemon_species: t.Union[int, str]) -> Route:
        """Get a pokemon species by its ID or name."""
        return Route(endpoint=f"/pokemon-species/{pokemon_species}")

    @classmethod
    def get_stat_endpoints(cls) -> Route:
        """Get a list of stat endpoints."""
        return Route(endpoint="/stat", payload={"limit": 10000})

    @classmethod
    def get_stat(cls, stat: t.Union[int, str]) -> Route:
        """Get a stat by its ID or name."""
        return Route(endpoint=f"/stat/{stat}")

    @classmethod
    def get_type_endpoints(cls) -> Route:
        """Get a list of type endpoints."""
        return Route(endpoint="/type", payload={"limit": 10000})

    @classmethod
    def get_type(cls, type_: t.Union[int, str]) -> Route:
        """Get a type by its ID or name."""
        return Route(endpoint=f"/type/{type_}")



import typing as t

import attrs

from pokelance.constants import ShowdownEnum
from pokelance.models import BaseModel

__all__: t.Tuple[str, ...] = ("ShowdownSprites",)


@attrs.define(kw_only=True, slots=True)
class ShowdownSprites(BaseModel):
    """
    Showdown sprites.

    As of now pokeapi doesn't include showdown sprites in their API payload.
    But they are available in their sprites repository.

    Attributes
    ----------
    back_default: str
        The default back sprite.
    front_default: str
        The default front sprite.
    back_shiny: str
        The shiny back sprite.
    front_shiny: str
        The shiny front sprite.
    """

    front_default: str = attrs.field(factory=str)
    front_shiny: str = attrs.field(factory=str)
    back_default: str = attrs.field(factory=str)
    back_shiny: str = attrs.field(factory=str)

    @classmethod
    def from_id(cls, id_: int) -> "ShowdownSprites":
        return cls(
            front_default=str(ShowdownEnum["front_default".upper()]).format(id_),
            front_shiny=str(ShowdownEnum["front_shiny".upper()]).format(id_),
            back_shiny=str(ShowdownEnum["back_shiny".upper()]).format(id_),
            back_default=str(ShowdownEnum["back_default".upper()]).format(id_),
        )



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = (
    "ContestComboSet",
    "ContestComboDetail",
    "MoveFlavorText",
    "MoveMetaData",
    "MoveStatChange",
    "PastMoveStatValues",
)


@attrs.define(slots=True, kw_only=True)
class ContestComboDetail(BaseModel):
    """A contest combo detail resource.

    Attributes
    ----------
    use_before: NamedResource
        A detail of moves this move can be used before, i.e. leading into this move.
    use_after: NamedResource
        A detail of moves this move can be used after, i.e. result in this move being used.
    """

    use_before: t.List[NamedResource] = attrs.field(factory=list)
    use_after: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ContestComboDetail":
        return cls(
            use_before=[NamedResource.from_payload(i or {}) for i in (payload.get("use_before", []) or [])],
            use_after=[NamedResource.from_payload(i or {}) for i in (payload.get("use_after", []) or [])],
        )


@attrs.define(slots=True, kw_only=True)
class ContestComboSet(BaseModel):
    """A contest combo set resource.

    Attributes
    ----------
    normal: ContestComboDetail
        A detail of normal moves in a contest combo.
    super_: ContestComboDetail
        A detail of super moves in a contest combo.
    """

    normal: ContestComboDetail = attrs.field(factory=ContestComboDetail)
    super_: ContestComboDetail = attrs.field(factory=ContestComboDetail)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ContestComboSet":
        return cls(
            normal=ContestComboDetail.from_payload(payload.get("normal", {}) or {}),
            super_=ContestComboDetail.from_payload(payload.get("super", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class MoveFlavorText(BaseModel):
    """A move flavor text resource.

    Attributes
    ----------
    flavor_text: str
        The localized flavor text for an api resource in a specific language.
    language: NamedResource
        The language this name is in.
    version_group: NamedResource
        The version group that uses this flavor text.
    """

    flavor_text: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveFlavorText":
        return cls(
            flavor_text=payload.get("flavor_text", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class MoveMetaData(BaseModel):
    """A move meta data resource.

    Attributes
    ----------
    ailment: NamedResource
        The status ailment this move inflicts on its target.
    category: NamedResource
        The category of move this move falls under, e.g. damage or ailment.
    min_hits: int
        The minimum number of times this move hits. Null if it always only hits once.
    max_hits: int
        The maximum number of times this move hits. Null if it always only hits once.
    min_turns: int
        The minimum number of turns this move continues to take effect. Null if it always only lasts one turn.
    max_turns: int
        The maximum number of turns this move continues to take effect. Null if it always only lasts one turn.
    drain: int
        HP drain (if positive) or Recoil damage (if negative), in percent of damage done.
    healing: int
        The amount of hp gained by the attacking Pokemon, in percent of it's maximum HP.
    crit_rate: int
        Critical hit rate bonus.
    ailment_chance: int
        The likelihood this attack will cause an ailment.
    flinch_chance: int
        The likelihood this attack will cause the target Pokémon to flinch.
    stat_chance: int
        The likelihood this attack will cause a stat change in the target Pokémon.
    """

    ailment: NamedResource = attrs.field(factory=NamedResource)
    category: NamedResource = attrs.field(factory=NamedResource)
    min_hits: int = attrs.field(factory=int)
    max_hits: int = attrs.field(factory=int)
    min_turns: int = attrs.field(factory=int)
    max_turns: int = attrs.field(factory=int)
    drain: int = attrs.field(factory=int)
    healing: int = attrs.field(factory=int)
    crit_rate: int = attrs.field(factory=int)
    ailment_chance: int = attrs.field(factory=int)
    flinch_chance: int = attrs.field(factory=int)
    stat_chance: int = attrs.field(factory=int)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveMetaData":
        return cls(
            ailment=NamedResource.from_payload(payload.get("ailment", {}) or {}),
            category=NamedResource.from_payload(payload.get("category", {}) or {}),
            min_hits=payload.get("min_hits", 0) or 1,
            max_hits=payload.get("max_hits", 0) or 1,
            min_turns=payload.get("min_turns", 0) or 1,
            max_turns=payload.get("max_turns", 0) or 1,
            drain=payload.get("drain", 0),
            healing=payload.get("healing", 0),
            crit_rate=payload.get("crit_rate", 0),
            ailment_chance=payload.get("ailment_chance", 0),
            flinch_chance=payload.get("flinch_chance", 0),
            stat_chance=payload.get("stat_chance", 0),
        )


@attrs.define(slots=True, kw_only=True)
class MoveStatChange(BaseModel):
    change: int = attrs.field(factory=int)
    stat: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MoveStatChange":
        return cls(
            change=payload.get("change", 0),
            stat=NamedResource.from_payload(payload.get("stat", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class PastMoveStatValues(BaseModel):
    accuracy: int = attrs.field(factory=int)
    effect_chance: int = attrs.field(factory=int)
    power: int = attrs.field(factory=int)
    pp: int = attrs.field(factory=int)
    effect_entries: t.List[NamedResource] = attrs.field(factory=list)
    type: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PastMoveStatValues":
        return cls(
            accuracy=payload.get("accuracy", 0),
            effect_chance=payload.get("effect_chance", 0),
            power=payload.get("power", 0),
            pp=payload.get("pp", 0),
            effect_entries=[NamedResource.from_payload(i) for i in payload.get("effect_entries", [])],
            type=NamedResource.from_payload(payload.get("type", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )



import enum
import typing as t

from pokelance.http import Endpoint
from pokelance.models import Language

__all__: t.Tuple[str, ...] = ("Languages",)
LANGUAGE = Endpoint.get_language_endpoints().url


class Languages(enum.Enum):
    """
    Languages that are available on Pokeapi.

    Attributes
    ----------
    JAPANESE: Language
        The Japanese language.
    ROOMAJI: Language
        The Japanese language written in roomaji.
    KOREAN: Language
        The Korean language.
    CHINESE: Language
        The Chinese language.
    FRENCH: Language
        The French language.
    GERMAN: Language
        The German language.
    SPANISH: Language
        The Spanish language.
    ITALIAN: Language
        The Italian language.
    CZECH: Language
        The Czech language.
    ENGLISH: Language
        The English language.
    JA: Language
        The Japanese language.
    CHINESE_SIMPLIFIED: Language
        The Chinese language written in simplified Chinese.
    PORTUGAL_BRAZILIAN: Language
        The Portuguese language written in Brazilian Portuguese.
    """

    JAPANESE = Language.from_payload(
        {
            "id": 1,
            "iso3166": "jp",
            "iso639": "ja",
            "name": "ja-Hrkt",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "日本語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "일본어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Japonais"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Japanisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Japonés"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Japanese"},
            ],
            "official": True,
        }
    )
    ROOMAJI = Language.from_payload(
        {
            "id": 2,
            "iso3166": "jp",
            "iso639": "ja",
            "name": "roomaji",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "正式ローマジ"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "정식 로마자"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Romaji"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Rōmaji"},
                {
                    "language": {"name": "en", "url": f"{LANGUAGE}/9/"},
                    "name": "Official roomaji",
                },
            ],
            "official": True,
        }
    )
    KOREAN = Language.from_payload(
        {
            "id": 3,
            "iso3166": "kr",
            "iso639": "ko",
            "name": "ko",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "韓国語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "한국어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Coréen"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Koreanisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Coreano"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Korean"},
            ],
            "official": True,
        }
    )
    CHINESE = Language.from_payload(
        {
            "id": 4,
            "iso3166": "cn",
            "iso639": "zh",
            "name": "zh-Hant",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "中国語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "중국어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Chinois"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Chinesisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Chino"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Chinese"},
            ],
            "official": True,
        }
    )
    FRENCH = Language.from_payload(
        {
            "id": 5,
            "iso3166": "fr",
            "iso639": "fr",
            "name": "fr",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "フランス語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "프랑스어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Français"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Französisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Francés"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "French"},
            ],
            "official": True,
        }
    )
    GERMAN = Language.from_payload(
        {
            "id": 6,
            "iso3166": "de",
            "iso639": "de",
            "name": "de",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "ドイツ語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "도이치어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Allemand"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Deutsch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Alemán"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "German"},
            ],
            "official": True,
        }
    )
    SPANISH = Language.from_payload(
        {
            "id": 7,
            "iso3166": "es",
            "iso639": "es",
            "name": "es",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "西語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "스페인어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Espagnol"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Spanisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Español"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Spanish"},
            ],
            "official": True,
        }
    )
    ITALIAN = Language.from_payload(
        {
            "id": 8,
            "iso3166": "it",
            "iso639": "it",
            "name": "it",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "伊語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "이탈리아어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Italien"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Italienisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Italiano"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Italian"},
            ],
            "official": True,
        }
    )
    ENGLISH = Language.from_payload(
        {
            "id": 9,
            "iso3166": "us",
            "iso639": "en",
            "name": "en",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "英語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "영어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Anglais"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Englisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Inglés"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "English"},
            ],
            "official": True,
        }
    )
    CZECH = Language.from_payload(
        {
            "id": 10,
            "iso3166": "cz",
            "iso639": "cs",
            "name": "cs",
            "names": [
                {"language": {"name": "ja-Hrkt", "url": f"{LANGUAGE}/1/"}, "name": "チェコ語"},
                {"language": {"name": "ko", "url": f"{LANGUAGE}/3/"}, "name": "체코어"},
                {"language": {"name": "fr", "url": f"{LANGUAGE}/5/"}, "name": "Tchèque"},
                {"language": {"name": "de", "url": f"{LANGUAGE}/6/"}, "name": "Tschechisch"},
                {"language": {"name": "es", "url": f"{LANGUAGE}/7/"}, "name": "Checo"},
                {"language": {"name": "en", "url": f"{LANGUAGE}/9/"}, "name": "Czech"},
            ],
            "official": False,
        }
    )
    JA = Language.from_payload({"id": 11, "iso3166": "jp", "iso639": "ja", "name": "ja", "names": [], "official": True})
    CHINESE_SIMPLIFIED = Language.from_payload(
        {"id": 12, "iso3166": "cn", "iso639": "zh", "name": "zh-Hans", "names": [], "official": True}
    )
    PORTUGAL_BRAZILIAN = Language.from_payload(
        {"id": 13, "iso3166": "br", "iso639": "pt-BR", "name": "pt-BR", "names": [], "official": False}
    )

    def __str__(self) -> str:
        """
        Returns the name of the language.
        """
        return str(self.value.name)

    def __int__(self) -> int:
        """
        Returns the id of the language.
        """
        return int(self.value.id)



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = (
    "ChainLink",
    "EvolutionDetail",
)
GENDER_MAP: t.Dict[int, str] = {1: "Female", 2: "Male", 3: "Genderless"}


@attrs.define(slots=True, kw_only=True)
class EvolutionDetail(BaseModel):
    """An evolution detail resource.

    Attributes
    ----------
    item: NamedResource
        The item required to cause evolution this into Pokémon species.
    trigger: NamedResource
        The type of event that triggers evolution into this Pokémon species.
    gender: str
        Gender of the evolving Pokémon species must be in order to evolve.
    held_item: NamedResource
        The item the evolving Pokémon species must be holding during the evolution trigger event.
    known_move: NamedResource
        The move that must be known by the evolving Pokémon species during the evolution trigger event.
    known_move_type: NamedResource
        The evolving Pokémon species must know a move with this type during the evolution trigger event.
    location: NamedResource
        The location the evolution must be triggered at.
    min_affection: int
        The minimum required level of affection the evolving Pokémon species must have.
    min_beauty: int
        The minimum required level of beauty the evolving Pokémon species must have.
    min_happiness: int
        The minimum required level of happiness the evolving Pokémon species must have.
    min_level: int
        The minimum required level of the evolving Pokémon species.
    needs_overworld_rain: bool
        Whether or not it must be raining in the overworld to cause evolution this Pokémon species.
    party_species: NamedResource
        The specific Pokémon species that must be in the players party in order for the evolution to occur.
    party_type: NamedResource
        The player must have a Pokémon of this type in their party during the evolution trigger event.
    relative_physical_stats: int
        The required relation between the Pokémon's Attack and Defense stats.
    time_of_day: str
        The time of day the evolution must be triggered at.
    trade_species: NamedResource
        The specific Pokémon species that must be traded with the evolving Pokémon species.
    turn_upside_down: bool
        Whether or not the 3DS needs to be turned upside-down as this Pokémon levels up.
    """

    item: NamedResource = attrs.field(factory=NamedResource)
    trigger: NamedResource = attrs.field(factory=NamedResource)
    gender: str = attrs.field(factory=str)
    held_item: NamedResource = attrs.field(factory=NamedResource)
    known_move: NamedResource = attrs.field(factory=NamedResource)
    known_move_type: NamedResource = attrs.field(factory=NamedResource)
    location: NamedResource = attrs.field(factory=NamedResource)
    min_level: int = attrs.field(factory=int)
    min_happiness: int = attrs.field(factory=int)
    min_beauty: int = attrs.field(factory=int)
    min_affection: int = attrs.field(factory=int)
    needs_overworld_rain: bool = attrs.field(factory=bool)
    party_species: NamedResource = attrs.field(factory=NamedResource)
    party_type: NamedResource = attrs.field(factory=NamedResource)
    relative_physical_stats: int = attrs.field(factory=int)
    time_of_day: str = attrs.field(factory=str)
    trade_species: NamedResource = attrs.field(factory=NamedResource)
    turn_upside_down: bool = attrs.field(factory=bool)

    @property
    def simplified_details(self) -> t.Dict[str, t.Any]:
        simplified_details: t.Dict[str, t.Any] = {}
        for k, v in self.to_dict().items():
            if ((is_dict := isinstance(v, dict)) and v.get("name") and v.get("url")) or (not is_dict and v):
                simplified_details[k] = v
        return simplified_details

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EvolutionDetail":
        return cls(
            item=NamedResource.from_payload(payload.get("item", {}) or {}),
            trigger=NamedResource.from_payload(payload.get("trigger", {}) or {}),
            gender=GENDER_MAP.get(payload.get("gender", 0), ""),
            held_item=NamedResource.from_payload(payload.get("held_item", {}) or {}),
            known_move=NamedResource.from_payload(payload.get("known_move", {}) or {}),
            known_move_type=NamedResource.from_payload(payload.get("known_move_type", {}) or {}),
            location=NamedResource.from_payload(payload.get("location", {}) or {}),
            min_level=payload.get("min_level", 0),
            min_happiness=payload.get("min_happiness", 0),
            min_beauty=payload.get("min_beauty", 0),
            min_affection=payload.get("min_affection", 0),
            needs_overworld_rain=payload.get("needs_overworld_rain", False),
            party_species=NamedResource.from_payload(payload.get("party_species", {}) or {}),
            party_type=NamedResource.from_payload(payload.get("party_type", {}) or {}),
            relative_physical_stats=payload.get("relative_physical_stats", 0),
            time_of_day=payload.get("time_of_day", ""),
            trade_species=NamedResource.from_payload(payload.get("trade_species", {}) or {}),
            turn_upside_down=payload.get("turn_upside_down", False),
        )


@attrs.define(slots=True, kw_only=True)
class ChainLink(BaseModel):
    """A chain link resource.

    Attributes
    ----------
    is_baby: bool
        Whether or not this link is for a baby Pokémon.
    species: NamedResource
        The Pokémon species at this point in the evolution chain.
    evolution_details: t.List[EvolutionDetail]
        A list of details regarding the specific details of the referenced Pokémon species evolution.
    evolves_to: t.List[ChainLink]
        A list of chain links.
    """

    is_baby: bool = attrs.field(factory=bool)
    species: NamedResource = attrs.field(factory=NamedResource)
    evolution_details: t.List[EvolutionDetail] = attrs.field(factory=list)
    evolves_to: t.List["ChainLink"] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ChainLink":
        return cls(
            is_baby=payload.get("is_baby", False),
            species=NamedResource.from_payload(payload.get("species", {}) or {}),
            evolution_details=[EvolutionDetail.from_payload(detail) for detail in payload.get("evolution_details", [])],
            evolves_to=[ChainLink.from_payload(link) for link in payload.get("evolves_to", [])],
        )



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = (
    "ItemSprites",
    "ItemHolderPokemon",
    "ItemHolderPokemonVersionDetail",
)


@attrs.define(slots=True, kw_only=True)
class ItemSprites(BaseModel):
    """An item sprites resource.

    Attributes
    ----------
    default: str
        The default depiction of this item.
    """

    default: str = attrs.field(factory=str)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, str]) -> "ItemSprites":
        return cls(default=payload.get("default", ""))


@attrs.define(slots=True, kw_only=True)
class ItemHolderPokemonVersionDetail(BaseModel):
    """An item holder pokemon version detail resource.

    Attributes
    ----------
    rarity: int
        The chance of this Pokémon holding this item in this version.
    version: NamedResource
        The version that this item is held in by the Pokémon.
    """

    rarity: int = attrs.field(factory=int)
    version: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemHolderPokemonVersionDetail":
        return cls(
            rarity=payload.get("rarity", 0),
            version=NamedResource.from_payload(payload.get("version", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class ItemHolderPokemon(BaseModel):
    """An item holder pokemon resource.

    Attributes
    ----------
    pokemon: str
        The Pokémon that holds this item.
    version_details: t.List[ItemHolderPokemonVersionDetail]
        The details for the version that this item is held in by the Pokémon.
    """

    pokemon: NamedResource = attrs.field(factory=NamedResource)
    version_details: t.List["ItemHolderPokemonVersionDetail"] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ItemHolderPokemon":
        return cls(
            pokemon=NamedResource.from_payload(payload.get("pokemon", {}) or {}),
            version_details=[
                ItemHolderPokemonVersionDetail.from_payload(version_detail)
                for version_detail in payload.get("version_details", [])
            ],
        )



import typing as t

from pokelance.http import Endpoint
from pokelance.models import Location as LocationModel
from pokelance.models import LocationArea, PalParkArea, Region

from ._base import BaseExtension

if t.TYPE_CHECKING:
    from pokelance import PokeLance
    from pokelance.cache import Location as LocationCache


__all__: t.Tuple[str, ...] = (
    "setup",
    "Location",
)


class Location(BaseExtension):
    """
    Extension for location related endpoints.

    Attributes
    ----------
    cache: pokelance.cache.Location
        The cache for this extension.
    """

    cache: "LocationCache"

    def get_location(self, name: t.Union[str, int]) -> t.Optional[LocationModel]:
        """Gets a location from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the location.

        Returns
        -------
        typing.Optional[pokelance.models.Location]
            The location if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the item pocket is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> location = client.location.get_location(1)
        >>> location.name
        'canalave-city'
        """
        route = Endpoint.get_location(name)
        self._validate_resource(self.cache.location, name, route)
        return self.cache.location.get(route, None)

    async def fetch_location(self, name: t.Union[str, int]) -> LocationModel:
        """Fetches a location from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the location.

        Returns
        -------
        LocationModel
            The location if it exists in the API, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the location is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     location = await client.location.fetch_location(1)
        ...     print(location.name)
        ...     await client.close()
        >>> asyncio.run(main())
        canalave-city
        """
        route = Endpoint.get_location(name)
        self._validate_resource(self.cache.location, name, route)
        data = await self._client.request(route)
        return self.cache.location.setdefault(route, LocationModel.from_payload(data))

    def get_location_area(self, name: t.Union[str, int]) -> t.Optional[LocationArea]:
        """Gets a location area from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the location area.

        Returns
        -------
        typing.Optional[pokelance.models.LocationArea]
            The location area if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the location area is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> location_area = client.location.get_location_area(1)
        >>> location_area.name
        'canalave-city-area'
        """
        route = Endpoint.get_location_area(name)
        self._validate_resource(self.cache.location_area, name, route)
        return self.cache.location_area.get(route, None)

    async def fetch_location_area(self, name: t.Union[str, int]) -> LocationArea:
        """Fetches a location area from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the location area.

        Returns
        -------
        LocationArea
            The location area if it exists in the API, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the location area is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     location_area = await client.location.fetch_location_area(1)
        ...     print(location_area.name)
        ...     await client.close()
        >>> asyncio.run(main())
        canalave-city-area
        """
        route = Endpoint.get_location_area(name)
        self._validate_resource(self.cache.location_area, name, route)
        data = await self._client.request(route)
        return self.cache.location_area.setdefault(route, LocationArea.from_payload(data))

    def get_pal_park_area(self, name: t.Union[str, int]) -> t.Optional[PalParkArea]:
        """Gets a pal park area from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the pal park area.

        Returns
        -------
        typing.Optional[pokelance.models.PalParkArea]
            The pal park area if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the pal park area is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> pal_park_area = client.location.get_pal_park_area(1)
        >>> pal_park_area.name
        'forest'
        """
        route = Endpoint.get_pal_park_area(name)
        self._validate_resource(self.cache.pal_park_area, name, route)
        return self.cache.pal_park_area.get(route, None)

    async def fetch_pal_park_area(self, name: t.Union[str, int]) -> PalParkArea:
        """Fetches a pal park area from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the pal park area.

        Returns
        -------
        PalParkArea
            The pal park area if it exists in the API, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the pal park area is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     pal_park_area = await client.location.fetch_pal_park_area(1)
        ...     print(pal_park_area.name)
        ...     await client.close()
        >>> asyncio.run(main())
        forest
        """
        route = Endpoint.get_pal_park_area(name)
        self._validate_resource(self.cache.pal_park_area, name, route)
        data = await self._client.request(route)
        return self.cache.pal_park_area.setdefault(route, PalParkArea.from_payload(data))

    def get_region(self, name: t.Union[str, int]) -> t.Optional[Region]:
        """Gets a region from the cache.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the region.

        Returns
        -------
        typing.Optional[pokelance.models.Region]
            The region if it exists in the cache, else None.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the region is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> client = PokeLance()
        >>> region = client.location.get_region(1)
        >>> region.name
        'kanto'
        """
        route = Endpoint.get_region(name)
        self._validate_resource(self.cache.region, name, route)
        return self.cache.region.get(route, None)

    async def fetch_region(self, name: t.Union[str, int]) -> Region:
        """Fetches a region from the API.

        Parameters
        ----------
        name: typing.Union[str, int]
            The name or id of the region.

        Returns
        -------
        Region
            The region if it exists in the API, else raises ResourceNotFound.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The name or id of the region is invalid.

        Examples
        --------

        >>> from pokelance import PokeLance
        >>> import asyncio
        >>> client = PokeLance()
        >>> async def main() -> None:
        ...     region = await client.location.fetch_region(1)
        ...     print(region.name)
        ...     await client.close()
        >>> asyncio.run(main())
        kanto
        """
        route = Endpoint.get_region(name)
        self._validate_resource(self.cache.region, name, route)
        data = await self._client.request(route)
        return self.cache.region.setdefault(route, Region.from_payload(data))


def setup(lance: "PokeLance") -> None:
    """Sets up the location cog."""
    lance.add_extension("location", Location(lance.http))



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Name, NamedResource

from .utils import BerryFlavorMap, FlavorBerryMap

__all__: t.Tuple[str, ...] = (
    "Berry",
    "BerryFirmness",
    "BerryFlavor",
)


@attrs.define(slots=True, kw_only=True)
class Berry(BaseModel):
    """A berry resource.

    Attributes
    ----------
    id: int
        The identifier for this berry resource.
    name: str
        The name for this berry resource.
    growth_time: int
        Time it takes the tree to grow one stage, in hours. Berry trees go through
        four of these growth stages before they can be picked.
    max_harvest: int
        The maximum number of these berries that can grow on one tree in Generation
        IV.
    natural_gift_power: int
        The power of the move "Natural Gift" when used with this Berry.
    size: int
        Berries are actually items. This is the number of those items.
    smoothness: int
        The speed at which this Berry dries out the soil as it grows. A higher
        rate means the soil dries more quickly.
    soil_dryness: int
        The firmness of this berry, used in making Pokéblocks or Poffins.
    flavors: typing.List[BerryFlavorMap]
        A list of references to each flavor a berry can have and the potency of
        each of those flavors in regard to this berry.
    item: pokelance.models.common.NamedResource
        The item that corresponds to this berry.
    natural_gift_type: pokelance.models.common.NamedResource
        The type inherited by "Natural Gift" when used with this Berry.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    growth_time: int = attrs.field(factory=int)
    max_harvest: int = attrs.field(factory=int)
    natural_gift_power: int = attrs.field(factory=int)
    size: int = attrs.field(factory=int)
    smoothness: int = attrs.field(factory=int)
    soil_dryness: int = attrs.field(factory=int)
    firmness: NamedResource = attrs.field(factory=NamedResource)
    flavors: t.List[BerryFlavorMap] = attrs.field(factory=list)
    item: NamedResource = attrs.field(factory=NamedResource)
    natural_gift_type: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, data: t.Dict[str, t.Any]) -> "Berry":
        return cls(
            id=data.get("id", 0),
            name=data.get("name", ""),
            growth_time=data.get("growth_time", 0),
            max_harvest=data.get("max_harvest", 0),
            natural_gift_power=data.get("natural_gift_power", 0),
            size=data.get("size", 0),
            smoothness=data.get("smoothness", 0),
            soil_dryness=data.get("soil_dryness", 0),
            firmness=NamedResource.from_payload(data.get("firmness", {}) or {}),
            flavors=[BerryFlavorMap.from_payload(flavor) for flavor in data.get("flavors", [])],
            item=NamedResource.from_payload(data.get("item", {}) or {}),
            natural_gift_type=NamedResource.from_payload(data.get("natural_gift_type", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class BerryFirmness(BaseModel):
    """A berry firmness resource.

    Attributes
    ----------
    id: int
        The identifier for this berry firmness resource.
    name: str
        The name for this berry firmness resource.
    berries: typing.List[pokelance.models.common.NamedResource]
        A list of the berries with this firmness.
    names: typing.List[pokelance.models.common.Name]
        A list of the name of this berry firmness listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    berries: t.List[NamedResource] = attrs.field(factory=list)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, data: t.Dict[str, t.Any]) -> "BerryFirmness":
        return cls(
            id=data.get("id", 0),
            name=data.get("name", ""),
            berries=[NamedResource.from_payload(berry) for berry in data.get("berries", [])],
            names=[Name.from_payload(name) for name in data.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class BerryFlavor(BaseModel):
    """A berry flavor resource.

    Attributes
    ----------
    id: int
        The identifier for this berry flavor resource.
    name: str
        The name for this berry flavor resource.
    berries: typing.List[FlavorBerryMap]
        A list of the berries with this flavor.
    contest_type: pokelance.models.common.NamedResource
        The contest type that correlates with this berry flavor.
    names: typing.List[pokelance.models.common.Name]
        The name of this berry flavor listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    berries: t.List[FlavorBerryMap] = attrs.field(factory=list)
    contest_type: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, data: t.Dict[str, t.Any]) -> "BerryFlavor":
        return cls(
            id=data.get("id", 0),
            name=data.get("name", ""),
            berries=[FlavorBerryMap.from_payload(berry) for berry in data.get("berries", [])],
            contest_type=NamedResource.from_payload(data.get("contest_type", {}) or {}),
            names=[Name.from_payload(name) for name in data.get("names", [])],
        )



import asyncio
import importlib
import json
import pathlib
import typing as t

import aiofiles
import attrs

if t.TYPE_CHECKING:
    from pokelance import models  # noqa: F401
    from pokelance.http import HttpClient, Route  # noqa: F401
    from pokelance.models import BaseModel  # noqa: F401


__all__: t.Tuple[str, ...] = (
    "PokemonCache",
    "BaseCache",
    "BerryCache",
    "PalParkAreaCache",
    "GenderCache",
    "GamesPokedexCache",
    "GamesVersionCache",
    "GamesGenerationCache",
    "BerryFlavorCache",
    "BerryFirmnessCache",
    "ContestTypeCache",
    "PokeathlonStatCache",
    "PokemonSpeciesCache",
    "PokemonLocationAreaCache",
    "ContestEffectCache",
    "SuperContestEffectCache",
    "PokemonColorCache",
    "PokemonFormCache",
    "PokemonHabitatCache",
    "PokemonShapeCache",
    "GrowthRateCache",
    "NatureCache",
    "TypeCache",
    "GenderCache",
    "GamesVersionGroupCache",
    "EggGroupCache",
    "EvolutionChainCache",
    "EvolutionTriggerCache",
    "EncounterConditionCache",
    "EncounterConditionValueCache",
    "EncounterMethodCache",
    "ItemCache",
    "ItemAttributeCache",
    "ItemCategoryCache",
    "ItemFlingEffectCache",
    "ItemPocketCache",
    "LocationCache",
    "LocationAreaCache",
    "MachineCache",
    "MoveCache",
    "MoveAilmentCache",
    "MoveBattleStyleCache",
    "MoveCategoryCache",
    "MoveDamageClassCache",
    "MoveLearnMethodCache",
    "MoveTargetCache",
    "AbilityCache",
    "CharacteristicCache",
    "PokemonCache",
    "PokemonSpeciesCache",
    "PokemonFormCache",
    "NatureCache",
    "TypeCache",
    "GenderCache",
    "GrowthRateCache",
    "PokemonLocationAreaCache",
    "StatCache",
    "RegionCache",
)

_KT = t.TypeVar("_KT", bound="Route")
_VT = t.TypeVar("_VT", bound="BaseModel")
_T = t.TypeVar("_T")


@attrs.define(kw_only=True, slots=True, frozen=True)
class Endpoint:
    id: t.Union[str, int] = attrs.field(factory=str)
    url: str = attrs.field(factory=str)

    def __str__(self) -> str:
        return str(self.id)


class BaseCache(t.MutableMapping[_KT, _VT]):
    """Base class for all caches.

    Parameters
    ----------
    max_size: int
        The maximum size of the cache.

    Attributes
    ----------
    _max_size: int
        The maximum size of the cache.
    _cache: typing.Dict[_KT, _VT]
        The cache itself.
    _endpoints: typing.Dict[str, int]
        The endpoints that are cached.
    _endpoints_cached: bool
        Whether or not the endpoints are cached.

    Examples
    --------
    >>> import asyncio
    >>> from pokelance import PokeLance
    >>>
    >>> async def main():
    ...     client = PokeLance()
    ...     print(await client.ping())
    ...     await asyncio.sleep(5)  # Wait for all the endpoints to load automatically. If not just load them manually.
    ...     # from pokelance.http import Endpoint
    ...     # data = await client.http.request(Endpoint.get_berry_endpoints())
    ...     # client.berry._cache.load_documents(str(client.berry.__class__.__name__).lower(), "berry", data)
    ...     # print(client.berry.cache.berry.endpoints)
    ...     # await client.berry.cache.berry.load_all(client.http)
    ...     print(client.berry.cache.berry)
    ...     await client.berry.cache.berry.save('temp')  # Save the cache to a file.
    ...     await client.berry.cache.berry.load('temp')  # Load the cache from a file.
    ...     print(client.berry.cache.berry)
    ...     await client.close()
    >>>
    >>> asyncio.run(main())
    """

    def __init__(self, max_size: int = 100) -> None:
        self._max_size = max_size
        self._cache: t.Dict[_KT, _VT] = {}
        self._endpoints: t.Dict[str, Endpoint] = {}
        self._endpoints_cached: bool = False

    def __getitem__(self, key: _KT) -> _VT:
        self._cache[key] = self._cache.pop(key)
        return self._cache[key]

    def __setitem__(self, key: _KT, value: _VT) -> None:
        if key in self._cache:
            self._cache[key] = self._cache.pop(key)
        else:
            if len(self._cache) >= self._max_size:
                self._cache.pop(list(self._cache.keys())[0])
            self._cache[key] = value

    def __delitem__(self, key: _KT) -> None:
        del self._cache[key]

    def __len__(self) -> int:
        return len(self._cache)

    def __iter__(self) -> t.Iterator[_KT]:
        return iter(self._cache)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._cache})"

    def keys(self) -> t.KeysView[_KT]:
        return self._cache.keys()

    def values(self) -> t.ValuesView[_VT]:
        return self._cache.values()

    def setdefault(self, __key: _KT, __default: t.Any = ...) -> _VT:
        if __key not in self:
            self[__key] = __default
        return self[__key]

    def clear(self) -> None:
        self._cache.clear()

    def items(self) -> t.ItemsView[_KT, _VT]:
        return self._cache.items()

    def get(self, key: _KT, /, default: t.Union[_VT, _T, None] = None) -> t.Union[_VT, _T, None]:  # type: ignore
        if key in self:
            return self[key]
        dummy: t.Dict[str, str] = {str(v): k for k, v in self._endpoints.items()}
        alias = dummy.get(key.endpoint.split("/")[-1]) or self._endpoints.get(key.endpoint.split("/")[-1])
        if alias:
            for k in self.keys():
                if k.endpoint.split("/")[-1] == str(alias):
                    return self[k]
        return default

    def load_documents(self, data: t.List[t.Dict[str, str]]) -> None:
        """Load documents into the cache.

        Parameters
        ----------
        data: typing.List[typing.Dict[str, str]]
            The data to load.
        """
        for document in data:
            self._endpoints[document["name"]] = Endpoint(url=document["url"], id=int(document["url"].split("/")[-2]))
        self._endpoints_cached = True

    async def wait_until_ready(self) -> None:
        """Wait until the cache is ready."""
        while not self._endpoints_cached:
            await asyncio.sleep(0.5)

    async def save(self, path: str = ".") -> None:
        """Save the cache to a file.

        Parameters
        ----------
        path: str
            The path to save the cache to.
        """
        pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        dummy: t.Dict[str, t.Dict[str, t.Any]] = {k.endpoint: attrs.asdict(v) for k, v in self.items()}
        async with aiofiles.open(pathlib.Path(f"{path}/{self.__class__.__name__}.json"), "w") as f:
            await f.write(json.dumps(dummy, indent=4, ensure_ascii=False))

    async def load(self, path: str = ".") -> None:
        """Load the cache from a file.

        Parameters
        ----------
        path: str
            The path to load the cache from.
        """
        async with aiofiles.open(pathlib.Path(f"{path}/{self.__class__.__name__}.json"), "r") as f:
            data = json.loads(await f.read())
        route_model = importlib.import_module("pokelance.http").__dict__["Route"]
        value_type = str(self.__orig_bases__[0].__args__[1]).split(".")[-1]  # type: ignore
        model: "models.BaseModel" = importlib.import_module("pokelance.models").__dict__[value_type]
        for endpoint, info in data.items():
            route = route_model(endpoint=endpoint)
            self.setdefault(route, model.from_payload(info))

    async def load_all(self, client: "HttpClient") -> None:
        """Load all documents into the cache.

        Parameters
        ----------
        client: HttpClient
            The client to use to load the documents.
        """
        if not self._endpoints_cached:
            raise RuntimeError("The endpoints have not been cached yet.")
        client._client.logger.info(f"Loading {self.__class__.__name__}...")
        route_model = importlib.import_module("pokelance.http").__dict__["Route"]
        value_type = str(self.__orig_bases__[0].__args__[1]).split(".")[-1]  # type: ignore
        model: "models.BaseModel" = importlib.import_module("pokelance.models").__dict__[value_type]
        for endpoint in self._endpoints.values():
            route = route_model(endpoint=f"/{endpoint.url.strip('/').split('/')[-2]}/{str(endpoint)}")
            self.setdefault(route, model.from_payload(await client.request(route)))
        client._client.logger.info(f"Loaded {self.__class__.__name__}.")

    @property
    def endpoints(self) -> t.Dict[str, Endpoint]:
        """The endpoints that are cached.

        Returns
        -------
        typing.Dict[str, Endpoint]
            The endpoints that are cached.
        """
        return self._endpoints

    @property
    def cache(self) -> t.Dict[_KT, _VT]:
        """The cache itself.

        Returns
        -------
        typing.Dict[_KT, _VT]
            The cache itself.
        """
        return self._cache


class SecondaryTypeCache(BaseCache[_KT, _VT]):
    """A cache for secondary types with differing endpoints."""

    def load_documents(self, data: t.List[t.Dict[str, str]]) -> None:
        """Load documents into the cache. Endpoints are different for secondary types.

        Parameters
        ----------
        data: typing.List[typing.Dict[str, str]]
            The data to load.
        """
        for document in data:
            self._endpoints[document["url"].split("/")[-2]] = Endpoint(
                url=document["url"], id=int(document["url"].split("/")[-2])
            )
        self._endpoints_cached = True


class BerryCache(BaseCache["Route", "models.Berry"]):
    """A cache for berries."""


class BerryFirmnessCache(BaseCache["Route", "models.BerryFirmness"]):
    """A cache for berry firmnesses."""


class BerryFlavorCache(BaseCache["Route", "models.BerryFlavor"]):
    """A cache for berry flavors."""


class ContestTypeCache(BaseCache["Route", "models.ContestType"]):
    """A cache for contest types."""


class PokemonCache(BaseCache["Route", "models.Pokemon"]):
    """A cache for pokemon."""


class AbilityCache(BaseCache["Route", "models.Ability"]):
    """A cache for abilities."""


class EggGroupCache(BaseCache["Route", "models.EggGroup"]):
    """A cache for egg groups."""


class GenderCache(BaseCache["Route", "models.Gender"]):
    """A cache for genders."""


class GrowthRateCache(BaseCache["Route", "models.GrowthRate"]):
    """A cache for growth rates."""


class NatureCache(BaseCache["Route", "models.Nature"]):
    """A cache for natures."""


class PokeathlonStatCache(BaseCache["Route", "models.PokeathlonStat"]):
    """A cache for pokeathlon stats."""


class PokemonColorCache(BaseCache["Route", "models.PokemonColor"]):
    """A cache for pokemon colors."""


class PokemonFormCache(BaseCache["Route", "models.PokemonForm"]):
    """A cache for pokemon forms."""


class PokemonLocationAreaCache(BaseCache["Route", "models.LocationAreaEncounter"]):
    """A cache for pokemon location areas."""


class PokemonHabitatCache(BaseCache["Route", "models.PokemonHabitats"]):
    """A cache for pokemon habitats."""


class PokemonShapeCache(BaseCache["Route", "models.PokemonShape"]):
    """A cache for pokemon shapes."""


class PokemonSpeciesCache(BaseCache["Route", "models.PokemonSpecies"]):
    """A cache for pokemon species."""


class StatCache(BaseCache["Route", "models.Stat"]):
    """A cache for stats."""


class TypeCache(BaseCache["Route", "models.Type"]):
    """A cache for types."""


class EncounterMethodCache(BaseCache["Route", "models.EncounterMethod"]):
    """A cache for encounter methods."""


class EncounterConditionCache(BaseCache["Route", "models.EncounterCondition"]):
    """A cache for encounter conditions."""


class EncounterConditionValueCache(BaseCache["Route", "models.EncounterConditionValue"]):
    """A cache for encounter condition values."""


class EvolutionTriggerCache(BaseCache["Route", "models.EvolutionTrigger"]):
    """A cache for evolution triggers."""


class GamesGenerationCache(BaseCache["Route", "models.Generation"]):
    """A cache for games generations."""


class GamesPokedexCache(BaseCache["Route", "models.Pokedex"]):
    """A cache for games pokedexes."""


class GamesVersionCache(BaseCache["Route", "models.Version"]):
    """A cache for games versions."""


class GamesVersionGroupCache(BaseCache["Route", "models.VersionGroup"]):
    """A cache for games version groups."""


class ItemCache(BaseCache["Route", "models.Item"]):
    """A cache for items."""


class ItemAttributeCache(BaseCache["Route", "models.ItemAttribute"]):
    """A cache for item attributes."""


class ItemCategoryCache(BaseCache["Route", "models.ItemCategory"]):
    """A cache for item categories."""


class ItemFlingEffectCache(BaseCache["Route", "models.ItemFlingEffect"]):
    """A cache for item fling effects."""


class ItemPocketCache(BaseCache["Route", "models.ItemPocket"]):
    """A cache for item pockets."""


class LocationCache(BaseCache["Route", "models.Location"]):
    """A cache for locations."""


class LocationAreaCache(BaseCache["Route", "models.LocationArea"]):
    """A cache for location areas."""


class PalParkAreaCache(BaseCache["Route", "models.PalParkArea"]):
    """A cache for pal park areas."""


class RegionCache(BaseCache["Route", "models.Region"]):
    """A cache for regions."""


class MoveCache(BaseCache["Route", "models.Move"]):
    """A cache for moves."""


class MoveAilmentCache(BaseCache["Route", "models.MoveAilment"]):
    """A cache for move ailments."""


class MoveBattleStyleCache(BaseCache["Route", "models.MoveBattleStyle"]):
    """A cache for move battle styles."""


class MoveCategoryCache(BaseCache["Route", "models.MoveCategory"]):
    """A cache for move categories."""


class MoveDamageClassCache(BaseCache["Route", "models.MoveDamageClass"]):
    """A cache for move damage classes."""


class MoveLearnMethodCache(BaseCache["Route", "models.MoveLearnMethod"]):
    """A cache for move learn methods."""


class MoveTargetCache(BaseCache["Route", "models.MoveTarget"]):
    """A cache for move targets."""


class MachineCache(SecondaryTypeCache["Route", "models.Machine"]):
    """A cache for machines."""


class EvolutionChainCache(SecondaryTypeCache["Route", "models.EvolutionChain"]):
    """A cache for evolution chains."""


class CharacteristicCache(SecondaryTypeCache["Route", "models.Characteristic"]):
    """A cache for characteristics."""


class ContestEffectCache(SecondaryTypeCache["Route", "models.ContestEffect"]):
    """A cache for contest effects."""


class SuperContestEffectCache(SecondaryTypeCache["Route", "models.SuperContestEffect"]):
    """A cache for super contest effects."""



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = ("ContestName",)


@attrs.define(slots=True, kw_only=True)
class ContestName(BaseModel):
    """A contest name resource.

    Attributes
    ----------
    name: str
        The name for this contest.
    color: str
        The color associated with this contest's name.
    language: NamedResource
        The language that this name is in.
    """

    name: str = attrs.field(factory=str)
    color: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ContestName":
        return cls(
            name=payload.get("name", ""),
            color=payload.get("color", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
        )



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = (
    "BerryFlavorMap",
    "FlavorBerryMap",
)


@attrs.define(slots=True, kw_only=True)
class BerryFlavorMap(BaseModel):
    """Represents a berry flavor map.

    Attributes
    ----------
    potency: int
        The potency of the referenced flavor for this berry.
    flavor: NamedResource
        The referenced berry flavor.
    """

    potency: int = attrs.field(factory=int)
    flavor: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "BerryFlavorMap":
        return cls(
            potency=payload.get("potency", 0),
            flavor=NamedResource.from_payload(payload.get("flavor", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class FlavorBerryMap(BaseModel):
    """Represents a flavor berry map.

    Attributes
    ----------
    potency: int
        The potency of the referenced berry for this flavor.
    berry: NamedResource
        The referenced berry.
    """

    potency: int = attrs.field(factory=int)
    berry: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "FlavorBerryMap":
        return cls(
            potency=payload.get("potency", 0),
            berry=NamedResource.from_payload(payload.get("berry", {}) or {}),
        )



import typing as t
from functools import lru_cache
from pathlib import Path

from .constants import ExtensionEnum, ExtensionsL
from .http import HttpClient
from .logger import Logger

if t.TYPE_CHECKING:
    import logging
    from types import TracebackType

    import aiohttp

    from .ext import BaseExtension, Berry, Contest, Encounter, Evolution, Game, Item, Location, Machine, Move, Pokemon
    from .models import BaseModel  # noqa: F401


__all__: t.Tuple[str, ...] = ("PokeLance",)


BaseType = t.TypeVar("BaseType", bound="BaseModel")


class PokeLance:
    """
    Main class to interact with the PokeAPI.


    Attributes
    ----------
    _http : HttpClient
        The HTTP client used to make requests to the PokeAPI.
    _logger : typing.Union[logging.Logger, Logger]
        The logger used to log information about the client.
    _ext_tasks : t.List[t.Tuple[t.Callable[[], t.Coroutine[t.Any, t.Any, None]], str]]
        A list of coroutines to load extension data.
    cache_endpoints : bool
        Whether to cache endpoints. Defaults to True.
    EXTENSIONS : Path
        The path to the extensions directory.
    berry : Berry
        The berry extension.
    contest : Contest
        The contest extension.
    encounter : Encounter
        The encounter extension.
    evolution : Evolution
        The evolution extension.
    game : Game
        The game extension.
    item : Item
        The item extension.
    location : Location
        The location extension.
    machine : Machine
        The machine extension.
    move : Move
        The move extension.
    pokemon : Pokemon
        The pokemon extension.

    Examples
    --------

    >>> import pokelance
    >>> import asyncio
    >>> client = pokelance.PokeLance()
    >>> async def main() -> None:
    ...     print(await client.ping())
    ...     await client.close()
    >>> asyncio.run(main())
    """

    EXTENSIONS: Path = Path(__file__).parent / "ext"
    _logger: t.Union["logging.Logger", Logger]
    berry: "Berry"
    contest: "Contest"
    evolution: "Evolution"
    game: "Game"
    item: "Item"
    location: "Location"
    machine: "Machine"
    move: "Move"
    pokemon: "Pokemon"
    encounter: "Encounter"

    def __init__(
        self,
        *,
        image_cache_size: int = 128,
        cache_size: int = 100,
        logger: t.Optional["logging.Logger"] = None,
        file_logging: bool = False,
        cache_endpoints: bool = True,
        session: t.Optional["aiohttp.ClientSession"] = None,
    ) -> None:
        """
        Parameters
        ----------
        image_cache_size : int
            The size of the image cache. Defaults to 128.
        cache_size : int
            The size of the cache to use for the HTTP client.
        logger : typing.Optional[logging.Logger]
            The logger to use. If not provided, a new logger will be created.
        file_logging : bool
            Whether to log to a file. Defaults to False.
        session : typing.Optional[aiohttp.ClientSession]
            The session to use for the HTTP client. It is recommended to use the default.

        Returns
        -------
        PokeLance
            The client.
        """
        self._logger = logger or Logger(name="pokelance", file_logging=file_logging)
        self._http = HttpClient(client=self, session=session, cache_size=cache_size)
        self.cache_endpoints = cache_endpoints
        self._ext_tasks: t.List[t.Tuple[t.Callable[[], t.Coroutine[t.Any, t.Any, None]], str]] = []
        self._image_cache_size = image_cache_size
        lru_cache(maxsize=self._image_cache_size)(self.get_image_async)
        self.setup_hook()

    async def __aenter__(self) -> "PokeLance":
        return self

    async def __aexit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional["TracebackType"],
    ) -> None:
        self.logger.warning("Closing session!")
        await self._http.close()

    def setup_hook(self) -> None:
        """
        The setup hook to be called after the client is created.
        This is called automatically when the client is created.
        It is not recommended to call this manually.
        """
        self._logger.info(f"Using cache size: {self._http.cache.max_size}")
        for extension in self.EXTENSIONS.iterdir():
            if extension.is_file() and extension.suffix == ".py":
                if "_" not in extension.stem:
                    module = __import__(f"pokelance.ext.{extension.stem}", fromlist=["setup"])
                    module.setup(self)
        self._logger.info("Setup complete")

    def add_extension(self, name: str, extension: "BaseExtension") -> None:
        """
        Adds an extension to the client. This is called automatically when an extension is loaded.

        Parameters
        ----------
        name : str
            The name of the extension.
        extension : BaseExtension
            The extension to add.
        """
        self._ext_tasks.append((extension.setup, name))
        setattr(self, name, extension)

    async def ping(self) -> float:
        """
        Pings the PokeAPI and returns the latency.

        Returns
        -------
        float
            The latency of the PokeAPI.
        """
        return await self._http.ping()

    async def close(self) -> None:
        """
        Closes the client session. Recommended to use this when the client is no longer needed.
        Not needed if the client is used in a context manager.
        """
        self.logger.warning("Closing session!")
        await self._http.close()

    async def getch_data(
        self, ext: t.Union[ExtensionEnum, ExtensionsL, str], category: str, id_: t.Union[int, str]
    ) -> BaseType:
        """
        A getch method that looks up the cache for the data first then gets it from the API if it is not found.

        Parameters
        ----------
        ext : Union[ExtensionEnum, ExtensionsL, str]
            The extension to get the data from.
        category : str
            The category to get the data from.
        id_ : Union[int, str]
            The ID of the data to get.

        Returns
        -------
        BaseType
            The data.

        Raises
        ------
        ValueError
            If the extension or category is invalid.
        ResourceNotFound
            If the data is not found.

        Examples
        --------
        >>> import pokelance
        >>> import asyncio
        >>> from pokelance.constants import ExtensionEnum
        >>> from pokelance.models import Pokemon
        >>> client = pokelance.PokeLance()
        >>> async def main() -> None:
        ...     pokemon: Pokemon = await client.getch_data(ExtensionEnum.Pokemon, "pokemon", 1)
        ...     print(pokemon.name)
        ...     await client.close()
        >>> asyncio.run(main())
        bulbasaur
        """
        if isinstance(ext, str) and (ext := str(ext).title()) not in ExtensionEnum.__members__:
            raise ValueError(f"Invalid extension: {ext}")
        categories = ExtensionEnum.get_categories(ext) if isinstance(ext, str) else ext.categories  # type: ignore
        if (category := category.lower().replace("-", "_")) not in categories:
            raise ValueError(f"Invalid category: {category}")
        ext_ = getattr(self, ext.lower()) if isinstance(ext, str) else getattr(self, ext.name.lower())
        get_, fetch_ = getattr(ext_, f"get_{category}"), getattr(ext_, f"fetch_{category}")
        return t.cast(BaseType, get_(id_) or await fetch_(id_))

    async def from_url(self, url: str) -> BaseType:
        """
        Constructs a request from urls present in the data.

        Parameters
        ----------
        url : str
            The URL to construct the request from.

        Returns
        -------
        BaseType
            The data from the URL.

        Raises
        ------
        ValueError
            If the url is invalid.
        """
        if params := ExtensionEnum.validate_url(url):
            return await self.getch_data(params.extension, params.category, params.value)
        raise ValueError(f"Invalid URL: {url}")

    async def get_image_async(self, url: str) -> bytes:
        """
        Gets an image from the url asynchronously.

        Parameters
        ----------
        url : str
            The URL to get the image from.

        Returns
        -------
        bytes
            The image data.
        """
        return await self._http.load_image(url)

    @property
    def ext_tasks(self) -> t.List[t.Tuple[t.Callable[[], t.Coroutine[t.Any, t.Any, None]], str]]:
        """
        A list of coroutines to load extension data.

        Returns
        -------
        typing.List[typing.Tuple[t.Callable[[], t.Coroutine[t.Any, t.Any, None]], str]]
            The list of tasks.
        """
        return self._ext_tasks

    @property
    def logger(self) -> t.Union["logging.Logger", Logger]:
        """
        The logger used to log information about the client.

        Returns
        -------
        logging.Logger
            The logger.
        """
        return self._logger

    @property
    def http(self) -> HttpClient:
        """
        The HTTP client used to make requests to the PokeAPI.

        Returns
        -------
        HttpClient
            The HTTP client.
        """
        return self._http

    @property
    def image_cache_size(self) -> int:
        """
        The size of the image cache.

        Returns
        -------
        int
            The size of the image cache.
        """
        return self._image_cache_size

    @image_cache_size.setter
    def image_cache_size(self, value: int) -> None:
        self._image_cache_size = value
        lru_cache(maxsize=self._image_cache_size)(self.get_image_async)



import typing as t
from difflib import get_close_matches

from pokelance.exceptions import ResourceNotFound
from pokelance.http import Endpoint

if t.TYPE_CHECKING:
    from pokelance.cache import BaseCache, Cache
    from pokelance.http import HttpClient, Route
    from pokelance.models import BaseModel


__all__: t.Tuple[str, ...] = ("BaseExtension",)
_KT = t.TypeVar("_KT", bound="Route")
_VT = t.TypeVar("_VT", bound="BaseModel")


class BaseExtension:
    """The base extension class.

    Parameters
    ----------
    client: pokelance.http.HttpClient
        The client to use for requests.

    Attributes
    ----------
    _client: pokelance.http.HttpClient
        The client to use for requests.
    _cache: pokelance.cache.Cache
        The cache to use for requests.
    """

    _cache: "Cache"

    def __init__(self, client: "HttpClient") -> None:
        """Initializes the extension.

        Parameters
        ----------
        client: pokelance.http.HttpClient
            The client to use for requests.

        Returns
        -------
        pokelance.ext.BaseExtension
            The extension.
        """
        self._client = client
        self._cache = self._client.cache
        self.cache = getattr(self._cache, self.__class__.__name__.lower())

    def _validate_resource(self, cache: "BaseCache[_KT, _VT]", resource: t.Union[str, int], route: "Route") -> None:
        """Validates a resource.

        Parameters
        ----------
        cache: pokelance.cache.BaseCache[typing.Any, typing.Any]
            The cache to use for the validation.
        resource: typing.Union[str, int]
            The resource to validate.
        route: pokelance.http.Route
            The route to use for the validation.

        Raises
        ------
        pokelance.exceptions.ResourceNotFound
            The resource was not found in the cache.
        """
        data: t.Set[str] = set(list(map(str, cache.endpoints.values())) + list(cache.endpoints.keys()))
        if data and str(resource) not in data:
            raise ResourceNotFound(self.get_message(str(resource), data), route, status=404)

    @staticmethod
    def get_message(case: str, data: t.Set[str]) -> str:
        """Gets the error message for a resource not found error.

        Parameters
        ----------
        case: str
            The case to use for the error message.
        data: typing.Set[str]
            The data to use for the error message.

        Returns
        -------
        str
            The error message.
        """
        matches = get_close_matches(case, data, n=10, cutoff=0.5)
        if matches:
            return f"Resource not found. Did you mean {', '.join(matches)}?"
        return "Resource not found."

    async def setup(self) -> None:
        """Sets up the extension."""
        for item in dir(self):
            if item.startswith("fetch_"):
                data = await self._client.request(
                    t.cast(t.Callable[[], "Route"], getattr(Endpoint, f"get_{item[6:]}_endpoints"))()
                )
                self._cache.load_documents(str(self.__class__.__name__), item[6:], data["results"])



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import Effect, FlavorText, NamedResource

from .utils import ContestName

__all__: t.Tuple[str, ...] = (
    "ContestType",
    "ContestEffect",
    "SuperContestEffect",
)


@attrs.define(slots=True, kw_only=True)
class ContestType(BaseModel):
    """A contest type resource.

    Attributes
    ----------
    id: int
        The identifier for this contest type resource.
    name: str
        The name for this contest type resource.
    berry_flavor: pokelance.models.common.NamedResource
        The berry flavor that correlates with this contest type.
    names: typing.List[pokelance.models.common.Name]
        The name of this contest type listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    berry_flavor: NamedResource = attrs.field(factory=NamedResource)
    names: t.List[ContestName] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ContestType":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            berry_flavor=NamedResource.from_payload(payload.get("berry_flavor", {}) or {}),
            names=[ContestName.from_payload(name) for name in payload.get("names", [])],
        )


@attrs.define(slots=True, kw_only=True)
class ContestEffect(BaseModel):
    """A contest effect resource.

    Attributes
    ----------
    id: int
        The identifier for this contest effect resource.
    appeal: int
        The base number of hearts the user of this move gets.
    jam: int
        The base number of hearts the user's opponent loses.
    effect_entries: typing.List[pokelance.models.common.Effect]
        The result of this contest effect listed in different languages.
    flavor_text_entries: typing.List[pokelance.models.common.FlavorText]
        The flavor text of this contest effect listed in different languages.
    """

    id: int = attrs.field(factory=int)
    appeal: int = attrs.field(factory=int)
    jam: int = attrs.field(factory=int)
    effect_entries: t.List[Effect] = attrs.field(factory=list)
    flavor_text_entries: t.List[FlavorText] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "ContestEffect":
        return cls(
            id=payload.get("id", 0),
            appeal=payload.get("appeal", 0),
            jam=payload.get("jam", 0),
            effect_entries=[Effect.from_payload(effect) for effect in payload.get("effect_entries", [])],
            flavor_text_entries=[
                FlavorText.from_payload(flavor_text) for flavor_text in payload.get("flavor_text_entries", [])
            ],
        )


@attrs.define(slots=True, kw_only=True)
class SuperContestEffect(BaseModel):
    """A super contest effect resource.

    Attributes
    ----------
    id: int
        The identifier for this super contest effect resource.
    appeal: int
        The level of appeal this super contest effect has.
    flavor_text_entries: typing.List[pokelance.models.common.FlavorText]
        The flavor text of this super contest effect listed in different languages.
    moves: typing.List[pokelance.models.common.NamedResource]
        A list of moves that have the effect when used in super contests.
    """

    id: int = attrs.field(factory=int)
    appeal: int = attrs.field(factory=int)
    flavor_text_entries: t.List[FlavorText] = attrs.field(factory=list)
    moves: t.List[NamedResource] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "SuperContestEffect":
        return cls(
            id=payload.get("id", 0),
            appeal=payload.get("appeal", 0),
            flavor_text_entries=[
                FlavorText.from_payload(flavor_text) for flavor_text in payload.get("flavor_text_entries", [])
            ],
            moves=[NamedResource.from_payload(move) for move in payload.get("moves", [])],
        )



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = (
    "EncounterVersionDetails",
    "EncounterMethodRate",
    "PokemonEncounter",
    "PalParkEncounterSpecies",
)


@attrs.define(slots=True, kw_only=True)
class EncounterVersionDetails(BaseModel):
    """An encounter version details resource.

    Attributes
    ----------
    rate: int
        The chance of an encounter to occur.
    version: NamedResource
        The version of this encounter.
    """

    rate: int = attrs.field(factory=int)
    version: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EncounterVersionDetails":
        return cls(
            rate=payload.get("rate", 0),
            version=NamedResource.from_payload(payload.get("version", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class EncounterMethodRate(BaseModel):
    """An encounter method rate resource.

    Attributes
    ----------
    encounter_method: NamedResource
        The method in which Pokémon may be encountered in an area.
    version_details: t.List[EncounterVersionDetails]
        A list of version details for the encounter.
    """

    encounter_method: NamedResource = attrs.field(factory=NamedResource)
    version_details: t.List[EncounterVersionDetails] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "EncounterMethodRate":
        return cls(
            encounter_method=NamedResource.from_payload(payload.get("encounter_method", {}) or {}),
            version_details=[EncounterVersionDetails.from_payload(i) for i in payload.get("version_details", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PokemonEncounter(BaseModel):
    """A pokemon encounter resource.

    Attributes
    ----------
    pokemon: NamedResource
        The Pokémon being encountered.
    version_details: t.List[EncounterVersionDetails]
        A list of versions and encounters with Pokémon that might happen.
    """

    pokemon: NamedResource = attrs.field(factory=NamedResource)
    version_details: t.List[EncounterVersionDetails] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PokemonEncounter":
        return cls(
            pokemon=NamedResource.from_payload(payload.get("pokemon", {}) or {}),
            version_details=[EncounterVersionDetails.from_payload(i) for i in payload.get("version_details", [])],
        )


@attrs.define(slots=True, kw_only=True)
class PalParkEncounterSpecies(BaseModel):
    """A pal park encounter species resource.

    Attributes
    ----------
    base_score: int
        The base score given to the player when this Pokémon is caught during a pal park run.
    rate: int
        The base rate for encountering this Pokémon in pal park.
    pokemon_species: NamedResource
        The Pokémon species being encountered.
    """

    base_score: int = attrs.field(factory=int)
    rate: int = attrs.field(factory=int)
    pokemon_species: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "PalParkEncounterSpecies":
        return cls(
            base_score=payload.get("base_score", 0),
            rate=payload.get("rate", 0),
            pokemon_species=NamedResource.from_payload(payload.get("pokemon_species", {}) or {}),
        )



import typing as t

from .cache_manager import (
    Base,
    BaseCache,
    Berry,
    Cache,
    Contest,
    Encounter,
    Evolution,
    Game,
    Item,
    Location,
    Machine,
    Move,
    Pokemon,
)

__all__: t.Tuple[str, ...] = (
    "Base",
    "Cache",
    "BaseCache",
    "Berry",
    "Contest",
    "Encounter",
    "Evolution",
    "Game",
    "Item",
    "Location",
    "Machine",
    "Move",
    "Pokemon",
)



import typing as t

import attrs

from pokelance.models import BaseModel

from .resources import NamedResource, Resource

__all__: t.Tuple[str, ...] = (
    "Description",
    "Effect",
    "Encounter",
    "FlavorText",
    "GenerationGameIndex",
    "MachineVersionDetail",
    "Name",
    "VerboseEffect",
    "VersionEncounterDetail",
    "VersionGameIndex",
    "VersionGroupFlavorText",
    "Language",
)


@attrs.define(slots=True, kw_only=True)
class Description(BaseModel):
    """Model for a description object

    Attributes
    ----------
    description: str
        The localized description for an API resource in a specific language.
    language: NamedResource
        The language this name is in.
    """

    description: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Description":
        return cls(
            description=payload.get("description", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class Effect(BaseModel):
    """Model for an effect object

    Attributes
    ----------
    effect: str
        The localized effect text for an API resource in a specific language.
    language: NamedResource
        The language this effect is in.
    """

    effect: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Effect":
        return cls(
            effect=payload.get("effect", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class Encounter(BaseModel):
    """Model for an encounter object

    Attributes
    ----------
    min_level: int
        The lowest level the Pokémon could be encountered at.
    max_level: int
        The highest level the Pokémon could be encountered at.
    condition_values: t.List[NamedResource]
        A list of condition values that must be in effect for this encounter to occur.
    chance: int
        The chance of the encounter to occur on a version of the game.
    method: NamedResource
        The method by which this encounter happens.
    """

    min_level: int = attrs.field(factory=int)
    max_level: int = attrs.field(factory=int)
    condition_values: t.List[Resource] = attrs.field(factory=list)
    chance: int = attrs.field(factory=int)
    method: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Encounter":
        return cls(
            min_level=payload.get("min_level", 0),
            max_level=payload.get("max_level", 0),
            condition_values=[Resource.from_payload(i) for i in payload.get("condition_values", [])],
            chance=payload.get("chance", 0),
            method=NamedResource.from_payload(payload.get("method", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class FlavorText(BaseModel):
    """Model for a flavor text object

    Attributes
    ----------
    flavor_text: str
        The localized flavor text for an API resource in a specific language.
    language: NamedResource
        The language this name is in.
    version: NamedResource
        The version that this flavor text is extracted from.
    """

    flavor_text: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)
    version: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "FlavorText":
        return cls(
            flavor_text=payload.get("flavor_text", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
            version=NamedResource.from_payload(payload.get("version", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class GenerationGameIndex(BaseModel):
    """Model for a generation game index object

    Attributes
    ----------
    game_index: int
        The internal id of an API resource within game data.
    generation: NamedResource
        The generation relevent to this game index.
    """

    game_index: int = attrs.field(factory=int)
    generation: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "GenerationGameIndex":
        return cls(
            game_index=payload.get("game_index", 0),
            generation=NamedResource.from_payload(payload.get("generation", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class MachineVersionDetail(BaseModel):
    """Model for a machine version detail object

    Attributes
    ----------
    machine: Resource
        The machine that teaches a move from an item.
    version_group: NamedResource
        The version group of this specific machine.
    """

    machine: Resource = attrs.field(factory=Resource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "MachineVersionDetail":
        return cls(
            machine=Resource.from_payload(payload.get("machine", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class Name(BaseModel):
    """Model for a name object

    Attributes
    ----------
    name: str
        The localized name for an API resource in a specific language.
    language: NamedResource
        The language this name is in.
    """

    name: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Name":
        return cls(name=payload.get("name", ""), language=NamedResource.from_payload(payload.get("language", {}) or {}))


@attrs.define(slots=True, kw_only=True)
class VerboseEffect(BaseModel):
    """Model for a verbose effect object

    Attributes
    ----------
    effect: str
        The localized effect text for an API resource in a specific language.
    short_effect: str
        The localized short effect text for an API resource in a specific language.
    language: NamedResource
        The language this effect is in.
    """

    effect: str = attrs.field(factory=str)
    short_effect: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "VerboseEffect":
        return cls(
            effect=payload.get("effect", ""),
            short_effect=payload.get("short_effect", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class VersionEncounterDetail(BaseModel):
    """Model for a version encounter detail object

    Attributes
    ----------
    max_chance: int
        The chance of an encounter to occur.
    encounter_details: t.List[Encounter]
        A list of encounters and their specifics.
    version: NamedResource
        The game version this encounter happens in.
    """

    version: NamedResource = attrs.field(factory=NamedResource)
    max_chance: int = attrs.field(factory=int)
    encounter_details: t.List[Encounter] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "VersionEncounterDetail":
        return cls(
            version=NamedResource.from_payload(payload.get("version", {}) or {}),
            max_chance=payload.get("max_chance", 0),
            encounter_details=[Encounter.from_payload(i) for i in payload.get("encounter_details", [])],
        )


@attrs.define(slots=True, kw_only=True)
class VersionGameIndex(BaseModel):
    """Model for a version game index object

    Attributes
    ----------
    game_index: int
        The internal id of an API resource within game data.
    version: NamedResource
        The version relevent to this game index.
    """

    game_index: int = attrs.field(factory=int)
    version: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "VersionGameIndex":
        return cls(
            game_index=payload.get("game_index", 0),
            version=NamedResource.from_payload(payload.get("version", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class VersionGroupFlavorText(BaseModel):
    """Model for a version group flavor text object

    Attributes
    ----------
    text: str
        The localized name for an API resource in a specific language.
    language: NamedResource
        The language this name is in.
    version_group: NamedResource
        The version group which uses this flavor text.
    """

    text: str = attrs.field(factory=str)
    language: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "VersionGroupFlavorText":
        return cls(
            text=payload.get("text", ""),
            language=NamedResource.from_payload(payload.get("language", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )


@attrs.define(slots=True, kw_only=True)
class Language(BaseModel):
    """Model for a language object

    Attributes
    ----------
    id: int
        The identifier for this resource.
    name: str
        The name for this resource.
    official: bool
        Whether or not the games are published in this language.
    iso639: str
        The two-letter code of the country where this language is spoken. Note that it is not unique.
    iso3166: str
        The two-letter code of the language. Note that it is not unique.
    names: t.List[Name]
        The name of this resource listed in different languages.
    """

    id: int = attrs.field(factory=int)
    name: str = attrs.field(factory=str)
    official: bool = attrs.field(factory=bool)
    iso639: str = attrs.field(factory=str)
    iso3166: str = attrs.field(factory=str)
    names: t.List[Name] = attrs.field(factory=list)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Language":
        return cls(
            id=payload.get("id", 0),
            name=payload.get("name", ""),
            official=payload.get("official", False),
            iso639=payload.get("iso639", ""),
            iso3166=payload.get("iso3166", ""),
            names=[Name.from_payload(i) for i in payload.get("names", [])],
        )



import typing as t

import attrs

from .cache import (
    AbilityCache,
    BaseCache,
    BerryCache,
    BerryFirmnessCache,
    BerryFlavorCache,
    CharacteristicCache,
    ContestEffectCache,
    ContestTypeCache,
    EggGroupCache,
    EncounterConditionCache,
    EncounterConditionValueCache,
    EncounterMethodCache,
    EvolutionChainCache,
    EvolutionTriggerCache,
    GamesGenerationCache,
    GamesPokedexCache,
    GamesVersionCache,
    GamesVersionGroupCache,
    GenderCache,
    GrowthRateCache,
    ItemAttributeCache,
    ItemCache,
    ItemCategoryCache,
    ItemFlingEffectCache,
    ItemPocketCache,
    LocationAreaCache,
    LocationCache,
    MachineCache,
    MoveAilmentCache,
    MoveBattleStyleCache,
    MoveCache,
    MoveCategoryCache,
    MoveDamageClassCache,
    MoveLearnMethodCache,
    MoveTargetCache,
    NatureCache,
    PalParkAreaCache,
    PokeathlonStatCache,
    PokemonCache,
    PokemonColorCache,
    PokemonFormCache,
    PokemonHabitatCache,
    PokemonLocationAreaCache,
    PokemonShapeCache,
    PokemonSpeciesCache,
    RegionCache,
    StatCache,
    SuperContestEffectCache,
    TypeCache,
)

__all__: t.Tuple[str, ...] = (
    "Cache",
    "Base",
    "BaseCache",
    "Berry",
    "Contest",
    "Encounter",
    "Evolution",
    "Game",
    "Item",
    "Location",
    "Machine",
    "Move",
    "Pokemon",
)


@attrs.define(slots=True, kw_only=True)
class Base:
    """Base class for all caches.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    """

    max_size: int

    def set_size(self, max_size: int = 100) -> None:
        """Set the maximum cache size.

        Parameters
        ----------
        max_size: int
            The maximum cache size.
        """
        self.max_size = max_size
        obj: attrs.Attribute[BaseCache[t.Any, t.Any]]
        for obj in self.__attrs_attrs__:  # type: ignore
            if isinstance(obj.default, BaseCache) and obj.default is not None:
                obj.default._max_size = max_size


@attrs.define(slots=True, kw_only=True)
class Encounter(Base):
    """Cache for encounter related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    encounter_method: EncounterMethodCache
        The method in which the encounter happens.
    encounter_condition: EncounterConditionCache
        The condition in which the encounter happens.
    encounter_condition_value: EncounterConditionValueCache
        The condition value in which the encounter happens.
    """

    max_size: int = 100
    encounter_method: EncounterMethodCache = attrs.field(default=EncounterMethodCache(max_size=max_size))
    encounter_condition: EncounterConditionCache = attrs.field(default=EncounterConditionCache(max_size=max_size))
    encounter_condition_value: EncounterConditionValueCache = attrs.field(
        default=EncounterConditionValueCache(max_size=max_size)
    )


@attrs.define(slots=True, kw_only=True)
class Evolution(Base):
    """Cache for evolution related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    evolution_chain: EvolutionChainCache
        The evolution chain of a Pokemon.
    evolution_trigger: EvolutionTriggerCache
        The trigger in which the evolution happens.
    """

    max_size: int = 100
    evolution_chain: EvolutionChainCache = attrs.field(default=EvolutionChainCache(max_size=max_size))
    evolution_trigger: EvolutionTriggerCache = attrs.field(default=EvolutionTriggerCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Machine(Base):
    """Cache for machine related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    machine: MachineCache
        The machine that teaches a move.
    """

    max_size: int = 100
    machine: MachineCache = attrs.field(default=MachineCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Game(Base):
    """Cache for game related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    generation: GamesGenerationCache
        The generation of a game.
    pokedex: GamesPokedexCache
        The pokedex of a game.
    version: GamesVersionCache
        The version of a game.
    version_group: GamesVersionGroupCache
        The version group of a game.
    """

    max_size: int = 100
    generation: GamesGenerationCache = attrs.field(default=GamesGenerationCache(max_size=max_size))
    pokedex: GamesPokedexCache = attrs.field(default=GamesPokedexCache(max_size=max_size))
    version: GamesVersionCache = attrs.field(default=GamesVersionCache(max_size=max_size))
    version_group: GamesVersionGroupCache = attrs.field(default=GamesVersionGroupCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Item(Base):
    """Cache for item related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    item: ItemCache
        The item.
    item_attribute: ItemAttributeCache
        The attribute of an item.
    item_category: ItemCategoryCache
        The category of an item.
    item_fling_effect: ItemFlingEffectCache
        The fling effect of an item.
    item_pocket: ItemPocketCache
        The pocket of an item.
    """

    max_size: int = 100
    item: ItemCache = attrs.field(default=ItemCache(max_size=max_size))
    item_attribute: ItemAttributeCache = attrs.field(default=ItemAttributeCache(max_size=max_size))
    item_category: ItemCategoryCache = attrs.field(default=ItemCategoryCache(max_size=max_size))
    item_fling_effect: ItemFlingEffectCache = attrs.field(default=ItemFlingEffectCache(max_size=max_size))
    item_pocket: ItemPocketCache = attrs.field(default=ItemPocketCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Location(Base):
    """Cache for location related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    location: LocationCache
        The location.
    location_area: LocationAreaCache
        The location area.
    """

    max_size: int = 100
    location: LocationCache = attrs.field(default=LocationCache(max_size=max_size))
    location_area: LocationAreaCache = attrs.field(default=LocationAreaCache(max_size=max_size))
    pal_park_area: PalParkAreaCache = attrs.field(default=PalParkAreaCache(max_size=max_size))
    region: RegionCache = attrs.field(default=RegionCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Move(Base):
    """Cache for move related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    move: MoveCache
        The move.
    move_ailment: MoveAilmentCache
        The ailment of a move.
    move_battle_style: MoveBattleStyleCache
        The battle style of a move.
    move_category: MoveCategoryCache
        The category of a move.
    move_damage_class: MoveDamageClassCache
        The damage class of a move.
    move_learn_method: MoveLearnMethodCache
        The learn method of a move.
    move_target: MoveTargetCache
        The target of a move.
    """

    max_size: int = 100
    move: MoveCache = attrs.field(default=MoveCache(max_size=max_size))
    move_ailment: MoveAilmentCache = attrs.field(default=MoveAilmentCache(max_size=max_size))
    move_battle_style: MoveBattleStyleCache = attrs.field(default=MoveBattleStyleCache(max_size=max_size))
    move_category: MoveCategoryCache = attrs.field(default=MoveCategoryCache(max_size=max_size))
    move_damage_class: MoveDamageClassCache = attrs.field(default=MoveDamageClassCache(max_size=max_size))
    move_learn_method: MoveLearnMethodCache = attrs.field(default=MoveLearnMethodCache(max_size=max_size))
    move_target: MoveTargetCache = attrs.field(default=MoveTargetCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Pokemon(Base):
    """Cache for pokemon related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    ability: AbilityCache
        The ability.
    characteristic: CharacteristicCache
        The characteristic.
    egg_group: EggGroupCache
        The egg group.
    gender: GenderCache
        The gender cache.
    growth_rate: GrowthRateCache
        The growth rate.
    nature: NatureCache
        The nature.
    pokeathlon_stat: PokeathlonStatCache
        The pokeathlon stat.
    pokemon: PokemonCache
        The pokemon.
    pokemon_color: PokemonColorCache
        The color of a pokemon.
    pokemon_form: PokemonFormCache
        The form of a pokemon.
    pokemon_habitat: PokemonHabitatCache
        The habitat of a pokemon.
    pokemon_shape: PokemonShapeCache
        The shape of a pokemon.
    pokemon_species: PokemonSpeciesCache
        The species of a pokemon.
    stat: StatCache
        The stat.
    type: TypeCache
        The type.
    """

    max_size: int = 100
    ability: AbilityCache = attrs.field(default=AbilityCache(max_size=max_size))
    characteristic: CharacteristicCache = attrs.field(default=CharacteristicCache(max_size=max_size))
    egg_group: EggGroupCache = attrs.field(default=EggGroupCache(max_size=max_size))
    gender: GenderCache = attrs.field(default=GenderCache(max_size=max_size))
    growth_rate: GrowthRateCache = attrs.field(default=GrowthRateCache(max_size=max_size))
    location_area_encounter: PokemonLocationAreaCache = attrs.field(default=PokemonLocationAreaCache(max_size=max_size))
    nature: NatureCache = attrs.field(default=NatureCache(max_size=max_size))
    pokeathlon_stat: PokeathlonStatCache = attrs.field(default=PokeathlonStatCache(max_size=max_size))
    pokemon: PokemonCache = attrs.field(default=PokemonCache(max_size=max_size))
    pokemon_color: PokemonColorCache = attrs.field(default=PokemonColorCache(max_size=max_size))
    pokemon_form: PokemonFormCache = attrs.field(default=PokemonFormCache(max_size=max_size))
    pokemon_habitat: PokemonHabitatCache = attrs.field(default=PokemonHabitatCache(max_size=max_size))
    pokemon_shape: PokemonShapeCache = attrs.field(default=PokemonShapeCache(max_size=max_size))
    pokemon_species: PokemonSpeciesCache = attrs.field(default=PokemonSpeciesCache(max_size=max_size))
    stat: StatCache = attrs.field(default=StatCache(max_size=max_size))
    type: TypeCache = attrs.field(default=TypeCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Contest(Base):
    """Cache for contest related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    contest_type: ContestTypeCache
        The contest type.
    contest_effect: ContestEffectCache
        The contest effect.
    super_contest_effect: SuperContestEffectCache
        The super contest effect.
    """

    max_size: int = 100
    contest_effect: ContestEffectCache = attrs.field(default=ContestEffectCache(max_size=max_size))
    contest_type: ContestTypeCache = attrs.field(default=ContestTypeCache(max_size=max_size))
    super_contest_effect: SuperContestEffectCache = attrs.field(default=SuperContestEffectCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Berry(Base):
    """Cache for berry related endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    berry: BerryCache
        The berry.
    berry_firmness: BerryFirmnessCache
        The berry firmness.
    berry_flavor: BerryFlavorCache
        The berry flavor.
    """

    max_size: int = 100
    berry: BerryCache = attrs.field(default=BerryCache(max_size=max_size))
    berry_firmness: BerryFirmnessCache = attrs.field(default=BerryFirmnessCache(max_size=max_size))
    berry_flavor: BerryFlavorCache = attrs.field(default=BerryFlavorCache(max_size=max_size))


@attrs.define(slots=True, kw_only=True)
class Cache:
    """Cache for all endpoints.

    Attributes
    ----------
    max_size: int
        The maximum cache size.
    berry: Berry
        The berry cache.
    contest: Contest
        The contest cache.
    encounter: Encounter
        The encounter cache.
    evolution: Evolution
        The evolution cache.
    game: Game
        The game cache.
    item: Item
        The item cache.
    location: Location
        The location cache.
    machine: Machine
        The machine cache.
    move: Move
        The move cache.
    pokemon: Pokemon
        The pokemon cache.
    """

    max_size: int = 100
    berry: Berry = attrs.field(default=Berry(max_size=max_size))
    contest: Contest = attrs.field(default=Contest(max_size=max_size))
    encounter: Encounter = attrs.field(default=Encounter(max_size=max_size))
    evolution: Evolution = attrs.field(default=Evolution(max_size=max_size))
    game: Game = attrs.field(default=Game(max_size=max_size))
    item: Item = attrs.field(default=Item(max_size=max_size))
    location: Location = attrs.field(default=Location(max_size=max_size))
    machine: Machine = attrs.field(default=Machine(max_size=max_size))
    move: Move = attrs.field(default=Move(max_size=max_size))
    pokemon: Pokemon = attrs.field(default=Pokemon(max_size=max_size))

    def __attrs_post_init__(self) -> None:
        obj: attrs.Attribute[Base]
        for obj in self.__attrs_attrs__:  # type: ignore
            if isinstance(obj.default, Base) and obj.default and obj.default.max_size != self.max_size:
                obj.default.set_size(self.max_size)

    def load_documents(self, category: str, _type: str, data: t.List[t.Dict[str, str]]) -> None:
        """Loads the endpoint data into the cache.

        Parameters
        ----------
        category: str
            The category of the endpoint.
        _type: str
            The type of the endpoint.
        data: typing.List[Dict[str, str]]
            The data to load.
        """
        getattr(getattr(self, category.lower()), _type).load_documents(data)



import typing as t

import attrs

from pokelance.models import BaseModel
from pokelance.models.common import NamedResource

__all__: t.Tuple[str, ...] = ("Machine",)


@attrs.define(slots=True, kw_only=True)
class Machine(BaseModel):
    """Machine model.

    Attributes
    ----------
    id: int
        The identifier for this resource.
    item: NamedResource
        The item that is required to use the TM or HM.
    move: NamedResource
        The move that is taught by the TM or HM.
    version_group: NamedResource
        The version group that this machine applies to.
    """

    id: int = attrs.field(factory=int)
    item: NamedResource = attrs.field(factory=NamedResource)
    move: NamedResource = attrs.field(factory=NamedResource)
    version_group: NamedResource = attrs.field(factory=NamedResource)

    @classmethod
    def from_payload(cls, payload: t.Dict[str, t.Any]) -> "Machine":
        return cls(
            id=payload.get("id", 0),
            item=NamedResource.from_payload(payload.get("item", {}) or {}),
            move=NamedResource.from_payload(payload.get("move", {}) or {}),
            version_group=NamedResource.from_payload(payload.get("version_group", {}) or {}),
        )



import asyncio
import time
import typing as t

import aiohttp

from pokelance.cache import Cache
from pokelance.exceptions import HTTPException, ImageNotFound

from .endpoints import Endpoint, Route

if t.TYPE_CHECKING:
    from pokelance import PokeLance


__all__: t.Tuple[str, ...] = (
    "HttpClient",
    "Route",
    "Endpoint",
)


@t.final
class HttpClient:
    """The HTTP client for PokeLance.

    Parameters
    ----------
    client: pokelance.PokeLance
        The client that this HTTP client is for.
    cache_size: int
        The size of the cache.
    session: aiohttp.ClientSession
        The session to use for the HTTP client.

    Attributes
    ----------
    session: aiohttp.ClientSession
        The session to use for the HTTP client.
    _is_ready: bool
        Whether the HTTP client is ready.
    _cache: pokelance.cache.Cache
        The cache to use for the HTTP client.
    _client: pokelance.PokeLance
        The client that this HTTP client is for.
    _tasks_queue: typing.List[asyncio.Task]
        The queue for the tasks.
    """

    __slots__: t.Tuple[str, ...] = (
        "_client",
        "session",
        "_cache",
        "_is_ready",
        "_tasks_queue",
    )

    def __init__(
        self, *, cache_size: int, client: "PokeLance", session: t.Optional[aiohttp.ClientSession] = None
    ) -> None:
        """Initializes the HTTP client.

        Parameters
        ----------
        cache_size: int
            The size of the cache.
        client: pokelance.PokeLance
            The client that this HTTP client is for.
        session: aiohttp.ClientSession
            The session to use for the HTTP client.

        Returns
        -------
        pokelance.http.HttpClient
            The HTTP client.
        """
        self._client = client
        self.session = session
        self._is_ready = False
        self._cache = Cache(max_size=cache_size)
        self._tasks_queue: t.List[asyncio.Task[None]] = []

    async def _load_ext(self, coroutine: t.Callable[[], t.Coroutine[t.Any, t.Any, None]], message: str) -> None:
        """
        Load an extension's resources.

        Parameters
        ----------
        coroutine: typing.Coroutine
            The coroutine to load.
        message: str
            The message to log.
        """
        task: t.Optional[asyncio.Task[None]] = asyncio.current_task()
        self._client.logger.debug(f"Loading {message}")
        await coroutine()
        if task is not None:
            self._tasks_queue.remove(task)
        self._client.logger.info(f"Loaded {message}")

    async def _schedule_tasks(self) -> None:
        """Schedules the tasks for the HTTP client."""
        total = len(self._client.ext_tasks)
        for num, (coroutine, name) in enumerate(self._client.ext_tasks):
            message = f"Extension {name} endpoints ({num + 1}/{total})"
            task = asyncio.create_task(coro=self._load_ext(coroutine, message), name=name)
            self._tasks_queue.append(task)
        self._client.ext_tasks.clear()

    async def close(self) -> None:
        """Closes the HTTP client."""
        for task in self._tasks_queue:
            if not task.done():
                task.cancel()
                self._client.logger.warning(f"Cancelled task {task.get_name()}")
        if self.session:
            await self.session.close()

    async def connect(self) -> None:
        """Connects the HTTP client."""
        self.session = self.session or aiohttp.ClientSession()
        if not self._is_ready:
            if self._client.cache_endpoints:
                await self._schedule_tasks()
            self._is_ready = True

    async def request(self, route: Route) -> t.Any:
        """Makes a request to the PokeAPI.

        Parameters
        ----------
        route: pokelance.http.Route
            The route to use for the request.

        Returns
        -------
        t.Any
            The response from the PokeAPI.

        Raises
        ------
        pokelance.exceptions.HTTPException
            An error occurred while making the request.
        """
        if self.session is None:
            await self.connect()
        if self.session is not None:
            async with self.session.request(route.method, route.url, params=route.payload) as response:
                if 300 > response.status >= 200:
                    self._client.logger.debug(f"Request to {route.url} was successful.")
                    return await response.json()
                else:
                    self._client.logger.error(f"Request to {route.url} was unsuccessful.")
                    raise HTTPException(str(response.reason), route, response.status).create()
        else:
            raise HTTPException("No session was provided.", route, -1).create()

    async def load_image(self, url: str) -> bytes:
        """Loads an image from the url.

        Parameters
        ----------
        url: str
            The URL to load the image from.

        Returns
        -------
        bytes
            The image.
        """
        if self.session is None:
            await self.connect()
        _image_formats = ("png", "jpg", "jpeg", "gif", "webp", "svg")
        if self.session is not None:
            async with self.session.get(url) as response:
                is_image = any(f_ in response.content_type for f_ in _image_formats)
                if 300 > response.status >= 200 and is_image:
                    self._client.logger.debug(f"Request to {url} was successful.")
                    return await response.read()
                else:
                    self._client.logger.error(f"Request to {url} was unsuccessful.")
                    message = f"Request to {url} was unsuccessful or the URL is not an image."
                    raise ImageNotFound(f"{message} ({response.content_type})", Route(), response.status)
        return b""

    async def ping(self) -> float:
        """Pings the PokeAPI and returns the latency.

        Returns
        -------
        float
            The latency of the PokeAPI.
        """
        start = time.perf_counter()
        await self.request(Route())
        return time.perf_counter() - start

    @property
    def cache(self) -> Cache:
        """The cache to use for the HTTP client.

        Returns
        -------
        pokelance.cache.Cache
            The cache.
        """
        return self._cache
